#------------------------------------------------------------------------------
#    Copyright (C) 2009 Florian Schanda <florian@schanda.org.uk>
#
#    This file is part of the Warzone Map Tools.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program. If not, see <http://www.gnu.org/licenses/>.
#------------------------------------------------------------------------------

""" The purose of this module is to provide a binding between clasp/gringo
    and python.

    The idea is to use compile(programs, ..., constants=values, ...)
    which will give you an object representing a grounded ASP program.

    This object provides the generator models(clasp_paramters, ...)
    yielding each model generated by clasp.
"""

import subprocess
import re
import os
import sys
import random

from itertools import imap

if not os.path.exists("potassco/gringo.bin") or \
        not os.path.exists("potassco/clasp.bin"):
    print "Error: Please compile clasp and gringo, included in potassco/"
    print "You can try the following:"
    print "  $ make"
    print "You will require cmake, re2c and cython for this."
    print
    print "If you already happen to have clasp and gringo installed,"
    print "then please create symlinks to them in potassco/ called"
    print "clasp.bin and gringo.bin."
    sys.exit(1)

def solutions(fd):
    """ Helper generator which reads from fd untit comes across a line
        starting with "Answer". It will then read the next two lines
        (the model and the optimistion) and yield them as a tuple.

        This continues until the end of file is reached.
    """
    while True:
        x = fd.readline()
        if x == "":
            return
        if x.startswith("Answer"):
            yield (fd.readline().strip(), fd.readline().strip())

def convert(x):
    """ Takes the string x and returns an int if possible; failing that a
        float; and failing that the original string.
    """
    try:
        return int(x)
    except ValueError:
        pass
    try:
        return float(x)
    except ValueError:
        pass
    return x

def parse(foo):
    """ Parse a statement in a model as follows:

        >>> parse("foo(1, 2, 3)")
        ('foo', [1, 2, 3])

        >>> parse("bar")
        ('bar', None)
    """
    tmp = foo.rstrip(")").split("(")
    if len(tmp) == 2:
        return (tmp[0], map(convert, tmp[1].split(",")))
    else:
        return (tmp[0], None)

class CompiledASP:
    """ This class represents a grounded ASP program. """
    def __init__(self, program):
        self.program = program
        
    def models(self, *clasp_parameters):
        """ Generator which will yield each model given by clasp.

            This function will also create the file debug.lp which
            contains the grounded program we are trying to solve.
        """
        args = ["potassco/clasp.bin"]
        if len(clasp_parameters) == 0:
            args += ["--rand-freq=0.05",
                     "--seed=%u" % random.getrandbits(31),
                     "--restart-on-model",
                     "--heuristic=BerkMin",
                     "--local-restart",
                     "--shuffle=1,1000"]
        else:
            args += ["--%s" % p for p in clasp_parameters]
        clasp = subprocess.Popen(args,
                                 stdin=subprocess.PIPE,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
        clasp.stdin.write(self.program)
        clasp.stdin.close()
        
        num_solutions = 0
        for s in solutions(clasp.stdout):
            if len(s[1]):
                print "clasp: solution with", s[1]
            rv = {}
            for k, v in map(parse, s[0].split()):
                if len(v) == 1:
                    rv[k] = rv.get(k, []) + v
                else:
                    rv[k] = rv.get(k, []) + [tuple(v)]
            rv["raw"] = s
            num_solutions += 1
            yield rv

        print "clasp: done."
        for l in clasp.stderr:
            print "err >", l.rstrip()

        print "ASP Return code:", clasp.wait()

        if num_solutions == 0:
            raise ValueError, ("Could not find a model within the current map"
                               " parameters. Please send the file"
                               " <tt>debug.lp</tt> to the developers.")

def compile(*programs, **constants):
    """ This function will concatenate a number of ASP programs, perform
        constant substitution and finally ground it. It returns an object
        of type CompiledASP.

        >>> tmp = compile("foo(1..x).\n", x=5)
        >>> for x in tmp.models(): print x["foo"]
        ...
        [1, 2, 3, 4, 5]
        clasp: done.
        ASP Return code: 10
    """
    params = {}
    tmp = ""
    for p in programs:
        if type(p) == file:
            fd = p
            p = fd.read()
            fd.close()
        for x in imap(lambda x: x.strip(), p.splitlines()):
            if x.startswith("#const"):
                k, v = x.split(" ", 1)[1].split("=")
                params[k] = convert(v.rstrip("."))
            else:
                tmp += "%s\n" % x
    for k, v in constants.iteritems():
        params[k] = v
    
    tmp = "\n".join(["#const %s=%s." % (k, v)
                     for k, v in params.iteritems()]) + tmp

    f = file("debug.lp", "w")
    f.write(tmp)
    f.close()

    gringo = subprocess.Popen(["potassco/gringo.bin"],
                              stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE)
    out, err = gringo.communicate(tmp)
    if gringo.returncode == 0:
        return CompiledASP(out)
    else:
        raise ValueError, err
