#!/usr/bin/env python
#------------------------------------------------------------------------------
#    Copyright (C) 2009 Florian Schanda <florian@schanda.org.uk>
#
#    This file is part of the Warzone Map Tools.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program. If not, see <http://www.gnu.org/licenses/>.
#------------------------------------------------------------------------------

from __future__ import division

# We import this here to get the error message about missing
# clasp.bin or gringo.bin as soon as possible
import asp

import sys
from PyQt4.QtCore import *
from PyQt4.QtGui import *

app = QApplication(sys.argv)
splash = QSplashScreen(QPixmap("img/diorama.jpg"))
splash.show()
app.processEvents()

from PyQt4.QtOpenGL import *
from OpenGL.GL import *
from OpenGL.GLU import *

import os
import random
import functools
import time

import mapwrite
import mapgen
import gen_cliff

from status_update import StatusUpdate

def nearest_pot(n):
    i = 1
    while i < n:
        i <<= 1
    return i

class MapDisplay(QGLWidget):
    def __init__(self, parent=None):
        QGLWidget.__init__(self, parent)
        self.texsize = None
        self.mapdata = None
        self.tex_preview = None
        self.scale = 3

    def initializeGL(self):
        glClearColor(0.0, 0.0, 0.5, 0.0)
        glEnable(GL_CULL_FACE)
        glEnable(GL_TEXTURE_2D)

        self.tex_preview = glGenTextures(1)

        glBindTexture(GL_TEXTURE_2D, self.tex_preview)
        glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
        glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP)
        glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP)

    def resizeGL(self, w, h):        
        glViewport(0, 0, w, h)

        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        glOrtho(0, w, h, 0, 0, 10)

    def paintGL(self):
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()
        glColor3d(1.0, 1.0, 1.0)

        if self.mapdata:
            glBindTexture(GL_TEXTURE_2D, self.tex_preview)
            glBegin(GL_QUADS)
            glTexCoord2d(0.0, 0.0)
            glVertex2i(0, 0)
            
            glTexCoord2d(0.0, 1.0)
            glVertex2i(0, self.texsize[1]*self.scale)

            glTexCoord2d(1.0, 1.0)
            glVertex2i(self.texsize[0]*self.scale, self.texsize[1]*self.scale)
            
            glTexCoord2d(1.0, 0.0)
            glVertex2i(self.texsize[0]*self.scale, 0)            
            glEnd()

    def newmap(self, mapdata, mapparam, tp):
        self.mapdata = mapdata

        print "Creating preview..."

        self.makeCurrent()
        glClearColor(0.0, 0.0, 0.0, 0.0)
        
        self.texsize = map(nearest_pot, (self.mapdata["width"],
                                         self.mapdata["height"]))

        self.setMinimumSize(self.mapdata["width"]*self.scale,
                            self.mapdata["height"]*self.scale)

        p_wells = set(x["pos"]
                      for x in self.mapdata["features"]
                      if x["name"] == "OilResource")
        
        TEXSET_WATER = tp.ts.get_set("Water")
        MAPBORDER = 3
        img = []
        for y in xrange(self.texsize[1]):
            mby = (y < MAPBORDER or y >= self.mapdata["height"] - MAPBORDER)
            for x in xrange(self.texsize[0]):
                if x >= self.mapdata["width"] or y >= self.mapdata["height"]:
                    img += (0, 0, 0)
                else:
                    mb = mby | (x < MAPBORDER or
                                x >= self.mapdata["width"]-MAPBORDER)

                    c = (self.mapdata["terrain"][(x, y)]["height"],) * 3
                    c = [int(c[0] * 0.8),
                         int(c[1] * 0.7),
                         int(c[2] * 0.5)]

                    # Lakes
                    if self.mapdata["terrain"][(x, y)]["tex"] in TEXSET_WATER:
                        c = [64, 64, 192]
                    
                    # Oil wells
                    elif (x, y) in p_wells:
                        c = [128, 255, 128]

                    # Buildings
                    elif (x, y) in tp.building_tiles:
                        c[0] += 48
                        c[1] += 32
                        c[2] += 16

                    # Roads
                    elif (x, y) in tp.road_tiles:
                        c[0] -= 32
                        c[1] -= 32
                        c[2] -= 32
                    
                    # Cliffs
                    elif (x, y) in tp.cliff_tiles:
                        c[0] = 96 + c[0] // 2
                        c[1] = 48 + c[1] // 4
                        c[2] = 48 + c[2] // 4

                    # MAP BORDER
                    if mb:
                        c = map(lambda x: x // 2, c)

                    img += tuple(map(lambda x: mapgen.clamp(x, 0, 255), c))
        
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB,
                     self.texsize[0], self.texsize[1],
                     0, GL_RGB, GL_UNSIGNED_BYTE,
                     img)

        self.updateGL()

def mk_lookup(d):
    def foo(x):
        return d[str(x)]
    return foo

class Controls(QWidget):
    def __init__(self, parent=None):
        QWidget.__init__(self, parent)

        lo_main = QVBoxLayout()
        lo_atr = QGridLayout()
        lo_atr.setVerticalSpacing(1)
 
        self.mapattr = [
            { "id": "name",
              "name": "Name",
              "tip": "The name of the map.",
              "type": ("string",),
              "value": ("Diorama" + time.strftime("-%m%d-%H%M",
                                              time.localtime()))
              #"value": "Test"
              },

            { "id": "wzversion",
              "name": "Map Version/Type",
              "tip": "The version of warzone to cater for.",
              "type": ("keyed_enum",
                       ("Optimised for 2.2", '2.2'),
                       ("Trunk only", "trunk"),
                       ),
              "value": "2.2"
              },

            { "id": "tileset",
              "name": "Tileset",
              "tip": "The tileset of the map.",
              "type": ("enum", "Random", "Arizona", "Urban", "Rockies"),
              "value": "Random"
              },

            { "id": "players",
              "name": "Players",
              "tip": "The number of players.",
              "type": ("intenum", 2, 4, 8),
              "value": 2
              },

            { "id": "mapstyle",
              "name": "Map Style",
              "tip": ("The overall map style.<br>"
                      "<b>Random</b> - An asymetric random map.<br>"
                      "<b>Symmetrical</b> - A X/Y symmetric map."
                      " Use this to quickly generate large 8 player maps,"
                      " as the total size will be 2*size - 1."),
              "type": ("keyed_enum",
                       ("Random", 'random'),
                       ("Symmetrical", "symmetry"),
                       ),
              "value": "random"
              },

            { "id": "global_feature",
              "name": "Global Feature",
              "type": ("keyed_enum",
                       ("None", 'none'),
                       ("Hotspot", "hotspot"),
                       ("Wasteland", "wasteland"),
                       ("Lake", "lake"),
                       ("Continent", "border_water"),
                       ),
              "value": "none"
              },

            { "id": "causeways",
              "name": "Causeways/Canyons",
              "tip": ("Influnce the numbers of ridges or canyons created.<br>"
                      "<b>Few</b> - This will create more open maps.<br>"
                      "<b>Lots</b> - This will create many ridges/canyons.<br>"
                      "<b>Don't care</b> - This will not impose any constrains"
                      " on the map generated."),
              "type": ("keyed_enum",
                       ("Few (3%)",    3),
                       ("Lots (75%)",  75),
                       ("Don't care",  0),
                       ),
              "value": 0
              },

            { "id": "basestyle",
              "name": "Base Defensibility",
              "tip": ("Constrain the number of cliffs directly around a"
                      " player's base."),
              "type": ("keyed_enum",
                       ("Easy (3)",   (1, 1, 2, 2)),
                       ("Medium (2)", (2, 2, 2, 2)),
                       ("Hard (1)", (3, 3, 3, 4)),
                       ("Extreme (0)", (4, 4, 3, 4)),
                       ("Don't care", (1, 4, 2, 4)),
                       ),
              "value": (1, 4, 2, 4)
              },

            { "id": "baseheight",
              "name": "Base Height",
              "tip": ("Player bases will be higher or lower than the"
                      " surrounding the terrain."),
              "type": ("keyed_enum",
                       ("Don't care", (0, 0)),
                       ("Sunken",     (1, 0)),
                       ("Raised",     (0, 1))
                       ),
              "value": (0, 0)
              },

            { "id": "baseoil",
              "name": "Player Oil Wells",
              "tip": ("Number of oil wells inside each player's base."),
              "type": ("int", 4, 8),
              "value": 6
              },

            { "id": "freeoil_clusters",
              "name": "Oil Clusters",
              "tip": ("Number of oil well clusters on the map which are"
                      " up for grabs. These are spaced roughly equally away"
                      " from each player base and each other."
                      "<br>For technical"
                      " reasons, this cannot be zero. To turn off non-player"
                      " oil wells, set this to one and the cluster size below"
                      " to zero."),
              "type": ("int", 1, 12),
              "value": 4
              },

            { "id": "freeoil_clustersize",
              "name": "Oil Cluster Size",
              "tip": ("Number of oil wells per cluster."),
              "type": ("int", 0, 4),
              "value": 2
              },

            { "id": "freeoil_style",
              "name": "Oil Defensibility",
              "tip": ("Number of cliffs around each cluster."),
              "type": ("keyed_enum",
                       ("Simple (3)", (1, 1)),
                       ("Easy (2-3)", (1, 2)),
                       ("Medium (1-2)", (2, 3)),
                       ("Difficult (0-1)", (3, 4)),
                       ("Hard (0)", (4, 4)),
                       ("Don't care", (1, 4))
                       ),
              "value": (1, 4)
              },

            { "id": "sx",
              "name": "Map Cell Width",
              "tip": ("This multiplied with cell scale below defines how"
                      " wide the map is."),
              "type": ("int", 6, 16),
              "value": 11
              },

            { "id": "sy",
              "name": "Map Cell Height",
              "tip": ("This multiplied with cell scale below defines how"
                      " high the map is."),
              "type": ("int", 6, 16),
              "value": 11
              },

            { "id": "scale",
              "name": "Cell Scale",
              "tip": ("Large values yield large maps, smaller values yield"
                      " more intricate maps."),
              "type": ("int", 8, 16),
              "value": 9
              },

            { "id": "depth",
              "name": "Depth",
              "tip": ("Number of distinct height levels."),
              "type": ("int", 3, 8),
              "value": 6
              },

            { "id": "density",
              "name": "Density (in %)",
              "tip": ("Higher values yield more cliffs, lower values yield"
                      " a more open terrain."),
              "type": ("float", 10.0, 75.0, 5.0),
              "value": 50.0
              },

            { "id": "waterdensity",
              "name": "Water Proportion (in %)",
              "tip": ("Gives the percentage of the map covered in water."),
              "type": ("float", 0.0, 75.0, 5.0),
              "value": 10.0
              },

            # Disabled, until we decide what to do with this
            #{ "id": "emptymiddle",
            #  "name": "Empty middle",
            #  "type": ("bool",),
            #  "value": False
            #  },

            { "id": "everythingreachable",
              "name": "Everything is reachable",
              "tip": ("This ensures that every location of the map is "
                      " reachable.<br>"
                      "Turning this off will give you unclimbable"
                      " mountains and sinkholes."),
              "type": ("bool",),
              "value": True
              },

            { "id": "asp_bases",
              "name": "Better bases",
              "tip": ("Generate neatly laid out and complete player bases with"
                      " 2 research facilities and factories and a power plant"
                      " and HQ building.<br>"
                      "If this option is turned off, only"
                      " incomplete player bases are created (suitable for"
                      " playing with the no bases option)."),
              "type": ("bool",),
              "value": False
              },
            ]

        cur_row = 0
        for attr in self.mapattr:
            cur_col = 0
            w = None
            t, args = attr["type"][0], attr["type"][1:]

            if t not in ["bool"]:
                lo_atr.addWidget(QLabel(attr["name"]), cur_row, cur_col)
                cur_col += 1

            if t == "int":
                w = QSpinBox()
                w.setRange(args[0], args[1])
                w.setValue(attr["value"])
                self.connect(w, SIGNAL("valueChanged(int)"),
                             functools.partial(self.update_attr,
                                               attr["id"], int))
            elif t == "float":
                w = QDoubleSpinBox()
                w.setDecimals(2)
                w.setRange(args[0], args[1])
                w.setSingleStep(args[2])
                w.setValue(attr["value"])
                self.connect(w, SIGNAL("valueChanged(double)"),
                             functools.partial(self.update_attr,
                                               attr["id"], float))
            elif t == "intenum":
                w = QComboBox()
                w.addItems(map(str, args))
                w.setCurrentIndex(list(args).index(attr["value"]))
                self.connect(w, SIGNAL("currentIndexChanged(QString)"),
                             functools.partial(self.update_attr,
                                               attr["id"],
                                               lambda x: x.toInt()[0]))
            elif t == "enum":
                w = QComboBox()
                w.addItems(list(args))
                w.setCurrentIndex(list(args).index(attr["value"]))
                self.connect(w, SIGNAL("currentIndexChanged(QString)"),
                             functools.partial(self.update_attr,
                                               attr["id"],
                                               lambda x: str(x)))
            elif t == "keyed_enum":
                w = QComboBox()
                tmp = [x[0] for x in args]
                w.addItems(tmp)
                w.setCurrentIndex([x[1] for x in args].index(attr["value"]))
                self.connect(w, SIGNAL("currentIndexChanged(QString)"),
                             functools.partial(self.update_attr,
                                               attr["id"],
                                               mk_lookup(dict(args))))
            elif t == "bool":
                w = QCheckBox()
                w.setText(attr["name"])
                w.setCheckState(Qt.Checked if attr["value"] else Qt.Unchecked)
                self.connect(w, SIGNAL("stateChanged(int)"),
                             functools.partial(\
                        self.update_attr,
                        attr["id"],
                        lambda x: True if x == Qt.Checked else False))
            elif t == "string":
                w = QLineEdit()
                w.setText(attr["value"])
                self.connect(w, SIGNAL("textChanged(QString)"),
                             functools.partial(self.update_attr,
                                               attr["id"],
                                               lambda x: str(x)))

            if "tip" in attr:
                w.setToolTip(attr["tip"])

            attr["widget"] = w
            lo_atr.addWidget(w, cur_row, cur_col, 1, 2-cur_col)

            cur_row += 1


        self.btn_bake = QPushButton("Bake!", self)
        self.connect(self.btn_bake, SIGNAL("clicked()"), self.bake)

        self.btn_save = QPushButton("Save", self)
        self.btn_save.setEnabled(False)
        self.connect(self.btn_save, SIGNAL("clicked()"), self.save)
        
        lo_main.addLayout(lo_atr)
        lo_main.addStretch()
        lo_main.addWidget(self.btn_bake)
        lo_main.addWidget(self.btn_save)
        self.setLayout(lo_main)

    def get_mapattr(self):
        return dict((x["id"], x["value"]) for x in self.mapattr)

    def update_attr(self, id, conv, value):
        attr = [x for x in self.mapattr if x["id"] == id][0]
        attr["value"] = conv(value)

    def bake(self):
        self.btn_bake.setEnabled(False)
        self.btn_save.setEnabled(False)
        try:
            for tp in gen_cliff.cliffmap(self.get_mapattr()):
                if isinstance(tp, StatusUpdate):
                    self.parent().parent().statusBar().showMessage(tp.text)
                    app.processEvents()
                else:
                    self.btn_save.setEnabled(True)
                    break
            self.parent().parent().statusBar().showMessage("Done.")
            self.emit(SIGNAL("new_map(PyQt_PyObject)"), tp)
        except ValueError, e:
            QMessageBox.information(None, "Error in map creation", str(e))
            raise
        except Exception, e:
            t, v, tb = sys.exc_info()
            QMessageBox.critical(None, str(t), str(e))
            raise
        finally:
            self.btn_bake.setEnabled(True)

    def save(self):
        self.emit(SIGNAL("save_map()"))


class MainWindow(QMainWindow):
    def __init__(self, parent=None):
        QMainWindow.__init__(self, parent)
        self.resize(800, 400)
        self.setWindowTitle("Diorama")

        exit = QAction("Quit", self)
        exit.setShortcut("Ctrl+Q")
        exit.setStatusTip("Quit application")
        self.connect(exit, SIGNAL("triggered()"), SLOT("close()"))

        menubar = self.menuBar()
        mainmenu = menubar.addMenu("&Diorama")
        mainmenu.addAction(exit)

        self.statusBar().showMessage("Ready...")

        self.md = MapDisplay(self)
        self.setCentralWidget(self.md)

        self.controls = Controls(self)
        self.connect(self.controls, SIGNAL("new_map(PyQt_PyObject)"),
                     self.update_map)
        self.connect(self.controls, SIGNAL("save_map()"),
                     self.save_map)

        cnt = QDockWidget("Controls", self)
        cnt.setAllowedAreas(Qt.LeftDockWidgetArea | Qt.RightDockWidgetArea)
        cnt.setFeatures(QDockWidget.DockWidgetMovable)
        cnt.setWidget(self.controls)
        self.addDockWidget(Qt.LeftDockWidgetArea, cnt)

        self.show()

        self.mapdata = None

    def update_map(self, tp):
        attr = self.controls.get_mapattr()
        self.mapdata = tp.save(attr["name"])
        self.md.newmap(self.mapdata, attr, tp)

    def save_map(self):
        attr = self.controls.get_mapattr()
        if self.mapdata:
            self.mapdata["name"] = attr["name"]
            mapwrite.install_map(self.mapdata)


mw = MainWindow()
mw.show()
splash.finish(mw)
app.exec_()
