#------------------------------------------------------------------------------
#    Copyright (C) 2009 Florian Schanda <florian@schanda.org.uk>
#
#    This file is part of the Warzone Map Tools.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program. If not, see <http://www.gnu.org/licenses/>.
#------------------------------------------------------------------------------

from __future__ import division, with_statement

import math
import random
import operator
from itertools import ifilter, imap, chain
from functools import partial

import mapgen
import mapwrite
import tilesets
import asp
import debug

import route as croute

from status_update import StatusUpdate

MAPBORDER = 4

def show_maze(m, pb, o):
    """ Pretty print maze (m), also showing the location of player bases (pb)
        and oil wells (o).
    """
    DEPTH = max(m[p] for p in m) + 1
    sx = max(x[0] for x in m) + 1
    sy = max(x[1] for x in m) + 1
    print DEPTH, sx, sy

    HM = ["\033[1;%um%u\033[0m" % (30+x-1, x-1)
          for x in xrange(1, DEPTH)]

    if sx <= 16 and sy <= 16:
        tmp = "  "
        for x in xrange(sx):
            tmp += "%x"% x
        print tmp
    for y in xrange(sy):
        tmp = ""
        if sx <= 16 and sy <= 16:
            tmp += "%x " % y
        for x in xrange(sx):
            c = " "
            if (x, y) in pb:
                c = "\033[32m@\033[0m"
            elif (x, y) in o:
                c = "\033[34;1mi\033[0m"
            elif m[(x, y)] == 0:
                c = "\033[34;44;1m~\033[0m"
            else:
                c = HM[m[(x, y)]-1]
            tmp += c
        print tmp


def pick_random(s, size=1):
    """ Used for building placement. """
    assert(type(s) == set)
    possibles = [p
                 for p in s
                 if mapgen.building(size, size, base=p) <= s]
    rv = random.choice(possibles)
    space = mapgen.building(size, size, base=rv)
    s -= space | mapgen.border(space)
    return rv


def pick_first(s, size=1):
    """ Used for non-random building placement. """
    assert(type(s) == set)
    rv = min(p
             for p in s
             if mapgen.building(size, size, base=p) <= s)
    space = mapgen.building(size, size, base=rv)
    s -= space | mapgen.border(space)
    return rv


def flood_fill(outline, start):
    to_fill = set([start])
    rv = set()
    tl = (min(p[0] for p in outline), min(p[1] for p in outline))
    br = (max(p[0] for p in outline), max(p[1] for p in outline))

    def is_inside(p):
        return (p[0] >= tl[0] and
                p[0] <= br[0] and
                p[1] >= tl[1] and
                p[1] <= br[1])

    while to_fill:
        t = to_fill.pop()
        rv.add(t)
        to_fill |= set(filter(is_inside, mapgen.field(n=1, base=t)))
        to_fill -= outline
        to_fill -= rv

    return rv


def draw_edge(begin, end, is_horizontal):
    rv = set([begin, end])
    b, e = begin, end
    for i in xrange(2):
        if is_horizontal:
            b = mapgen.vec_add(b, ( 1, 0))
            e = mapgen.vec_add(e, (-1, 0))
        else:
            b = mapgen.vec_add(b, (0,  1))
            e = mapgen.vec_add(e, (0, -1))
        rv |= set([b, e])

    p = b
    diff = list(mapgen.vec_sub(e, b))
    while diff[0] != 0 or diff[1] != 0:
        total = reduce(operator.add, map(abs, diff))
        if random.randint(1, total) <= abs(diff[0]):
            p = mapgen.vec_add(p, (1, 0) if diff[0] > 0 else (-1, 0))
            diff[0] += 1 if diff[0] < 0 else -1
        else:
            p = mapgen.vec_add(p, (0, 1) if diff[1] > 0 else (0, -1))
            diff[1] += 1 if diff[1] < 0 else -1
        rv.add(p)

    return rv

def route(a, b, valid, cheap):
    assert(a in valid)
    assert(b in valid)
    return croute.route(a, b, valid, cheap)

#
# This is the code to use cliff_layout.lp and generate a useful map.
# Currently, this remainder of this file is effectively one spaghetti
# generator function.
#

def cliffmap(mattr):
    #--------------------------------------------------------------------------
    # Set up various constants and useful things
    #--------------------------------------------------------------------------
    #random.seed(0)

    sx, sy = mattr["sx"], mattr["sy"]
    ox, oy = sx, sy

    DEPTH = mattr["depth"]+1
    SCALE = mattr["scale"]
    DENSITY = mattr["density"]/100.0
    DENSITY_WATER = mattr["waterdensity"]/100.0
    DENSITY_CW = mattr["causeways"]/100.0
    NUM_PLAYERS = mattr["players"]
    NUM_PLAYER_OIL = mattr["baseoil"]

    NUM_FREE_OIL = mattr["freeoil_clusters"]
    OIL_MIN_FREE, OIL_MAX_FREE = mattr["freeoil_style"]
    FREE_OIL_CLUSTER_SIZE = mattr["freeoil_clustersize"]
    
    EMPTY_MIDDLE = mattr.get("emptymiddle", 0)
    EVERYTHING_REACHABLE = 1 if mattr["everythingreachable"] else 0
    BASE_N_MIN, BASE_N_MAX, BASE_E_MIN, BASE_E_MAX = mattr["basestyle"]
    BASE_SUNKEN, BASE_RAISED = mattr["baseheight"]

    # Additional boundary for base placement from map edges
    BASE_EB = 0

    GLOBAL_FEATURE = mattr.get("global_feature", "non")

    MAPSTYLE = mattr["mapstyle"]
    if MAPSTYLE == "symmetry":
        sx *= 2
        sy *= 2
        sx -= 1
        sy -= 1
        if ox != oy:
            raise ValueError, ("Symmetrical maps must be square. Please set"
                               " <b>map cell width</b> to be equal to <b>map"
                               " cell height</b>.")

    if NUM_PLAYERS == 8 and (sx*sy < 100):
        raise ValueError, ("Not enough space for 8 players - try increasing"
                           " either <b>map cell width</b> or <b>map cell"
                           " height</b> so that there are at least 100 cells.")

    ASP_FOR_BASES = mattr["asp_bases"]

    WZ_VERSION = mattr.get("wzversion", "2.2")
    if WZ_VERSION == "2.2": 
        tp = mapgen.TerrainPainter(sx*SCALE, sy*SCALE,
                                   tilesets.get_tileset(mattr["tileset"]))
    elif WZ_VERSION == "trunk":
        # Don't randomly select urban for trunk/, it sucks
        if mattr["tileset"] == "Random":
            mattr["tileset"] = random.choice(["Arizona", "Rockies"])

        tp = mapgen.TrunkTerrainPainter(sx*SCALE, sy*SCALE,
                                        tilesets.get_tileset(mattr["tileset"]))

    # This is a set of all valid tiles inside the map
    tiles_in_border = frozenset(filter(lambda p: (p[0] >= MAPBORDER and
                                                  p[1] >= MAPBORDER and
                                                  p[0] < sx*SCALE-MAPBORDER and
                                                  p[1] < sy*SCALE-MAPBORDER),
                                       tp.valid_tiles))


    yield StatusUpdate(text="Initialisation done.")
            
    #--------------------------------------------------------------------------
    # Maze
    #--------------------------------------------------------------------------

    NUM_WALLS = int(float(ox*oy) * DENSITY)
    NUM_WATER = int(float(ox*oy) * DENSITY_WATER)

    CW_CARE = int(DENSITY_CW > 0)
    NUM_CW = int(NUM_WALLS * DENSITY_CW * 0.5)
    NUM_CW_MIN = max(0, NUM_CW - 2) if CW_CARE else 0
    NUM_CW_MAX = NUM_CW + 2 if CW_CARE else 0

    NUM_WALLS_SYM = int(float(sx*sy) * DENSITY)
    NUM_WATER_SYM = int(float(sx*sy) * DENSITY_WATER)

    LP_GRID = file("grid.lp", "r").read()
    LP_CLIFF_LAYOUT = file("cliff-layout.lp", "r").read()
    LP_LAYOUT = file("layout.lp", "r").read()
    LP_RESOURCES = file("resources.lp", "r").read()
    LP_DISTANCE = file("distance.lp", "r").read()
    

    ASP_GF = ""
    if GLOBAL_FEATURE == "hotspot":
        if MAPSTYLE == "random":
            middle = mapgen.field(n=1, base=(sx//2, sy//2))
            for i in xrange(sy):
                if (sx//2, i) in middle:
                    continue
                ASP_GF += "cellLevel(%u, %u, %u).\n" % (sx//2, i, DEPTH-1)
                ASP_GF += ":- cellLevel(%u, %u, %u).\n" % (sx//2-1, i, DEPTH-1)
                ASP_GF += ":- cellLevel(%u, %u, %u).\n" % (sx//2+1, i, DEPTH-1)
                ASP_GF += ":- cellLevel(%u, %u, %u).\n" % (sx//2-1, i, DEPTH-2)
                ASP_GF += ":- cellLevel(%u, %u, %u).\n" % (sx//2+1, i, DEPTH-2)
            for i in xrange(sx):
                if (i, sy//2) in middle:
                    continue
                ASP_GF += "cellLevel(%u, %u, %u).\n" % (i, sy//2, DEPTH-1)
                ASP_GF += ":- cellLevel(%u, %u, %u).\n" % (i, sy//2-1, DEPTH-1)
                ASP_GF += ":- cellLevel(%u, %u, %u).\n" % (i, sy//2+1, DEPTH-1)
                ASP_GF += ":- cellLevel(%u, %u, %u).\n" % (i, sy//2-1, DEPTH-2)
                ASP_GF += ":- cellLevel(%u, %u, %u).\n" % (i, sy//2+1, DEPTH-2)
        elif MAPSTYLE == "symmetry":
            for i in xrange(oy-2):
                ASP_GF += "cellLevel(%u, %u, %u).\n" % (ox-1, i, DEPTH-1)
                ASP_GF += ":- cellLevel(%u, %u, %u).\n" % (ox-2, i, DEPTH-1)
                ASP_GF += ":- cellLevel(%u, %u, %u).\n" % (ox-2, i, DEPTH-2)
            for i in xrange(ox-2):
                ASP_GF += "cellLevel(%u, %u, %u).\n" % (i, oy-1, DEPTH-1)
                ASP_GF += ":- cellLevel(%u, %u, %u).\n" % (i, oy-2, DEPTH-1)
                ASP_GF += ":- cellLevel(%u, %u, %u).\n" % (i, oy-2, DEPTH-2)
    elif GLOBAL_FEATURE == "wasteland":
        if MAPSTYLE == "random":
            nocliff = mapgen.square(3, 3, sx-5, sy-5)
        elif MAPSTYLE == "symmetry":
            nocliff = mapgen.square(ox-6, oy-6, ox-2, oy-2)
        for x, y in sorted(nocliff):
            ASP_GF += ":- cliffDown(%u, %u).\n" % (x, y)
            ASP_GF += ":- cliffRight(%u, %u).\n" % (x, y)
    elif GLOBAL_FEATURE == "lake":
        if MAPSTYLE == "random":
            lake = mapgen.square(4, 4, sx-5, sy-5)
        elif MAPSTYLE == "symmetry":
            lake = mapgen.square(ox-5, ox-5, ox-1, oy-1)
        lake |= mapgen.randomborder(lake, 25)
        for x, y in sorted(lake):
            ASP_GF += "cellLevel(%u, %u, 0).\n" % (x, y)
        NUM_WATER += len(lake)
    elif GLOBAL_FEATURE == "border_water":
        if MAPSTYLE == "random":
            lake = (mapgen.square(0, 0, sx-1, sy-1) -
                    mapgen.square(1, 1, sx-2, sy-2))
        elif MAPSTYLE == "symmetry":
            lake = (mapgen.square(0, 0, sx-1, sy-1) -
                    mapgen.square(1, 1, sx-1, sy-1))
        for x, y in sorted(lake):
            ASP_GF += "cellLevel(%u, %u, 0).\n" % (x, y)
        NUM_WATER += len(lake)
        BASE_EB += 1

    ASP_EXTRA_BOUNDARY = ""
    if BASE_EB:
        for i in xrange(1, BASE_EB+1):
            ASP_EXTRA_BOUNDARY += ":- baseLocation(B,X,%u).\n" % i
            ASP_EXTRA_BOUNDARY += ":- baseLocation(B,X,height-%u).\n" % (i+1)
            ASP_EXTRA_BOUNDARY += ":- baseLocation(B,%u,Y).\n" % i
            ASP_EXTRA_BOUNDARY += ":- baseLocation(B,width-%u,Y).\n" % (i+1)
            
    stdc = partial(asp.compile,
                   LP_GRID,
                   LP_CLIFF_LAYOUT,
                   LP_LAYOUT,
                   LP_RESOURCES,
                   LP_DISTANCE,
                   ASP_EXTRA_BOUNDARY,
                   depth=DEPTH,
                   isEverythingReachable=EVERYTHING_REACHABLE,
                   minOilNeighbours=OIL_MIN_FREE,
                   maxOilNeighbours=OIL_MAX_FREE,
                   minBaseNeighbours=BASE_N_MIN,
                   maxBaseNeighbours=BASE_N_MAX,
                   minEntranceNeighbours=BASE_E_MIN,
                   maxEntranceNeighbours=BASE_E_MAX,
                   isSunkenBase=BASE_SUNKEN,
                   isRaisedBase=BASE_RAISED,
                   isLowestLevelWater=1)

    if MAPSTYLE == "random":
        # This hint should make 2 and 4 player maps much faster
        ASP_BASEDEF = ""
        if NUM_PLAYERS == 2:
            ASP_BASEDEF += "baseLocation(1, %u, %u).\n" % (1+BASE_EB,
                                                           1+BASE_EB)
            ASP_BASEDEF += "baseLocation(2, %u, %u).\n" % (ox-2-BASE_EB,
                                                           oy-2-BASE_EB)
        elif NUM_PLAYERS == 4 and ox == oy:
            ASP_BASEDEF += "baseLocation(1, %u, %u).\n" % (1+BASE_EB,
                                                           1+BASE_EB)
            ASP_BASEDEF += "baseLocation(2, %u, %u).\n" % (ox-2-BASE_EB,
                                                           1+BASE_EB)
            ASP_BASEDEF += "baseLocation(3, %u, %u).\n" % (1+BASE_EB,
                                                           oy-2-BASE_EB)
            ASP_BASEDEF += "baseLocation(4, %u, %u).\n" % (ox-2-BASE_EB,
                                                           oy-2-BASE_EB)

        asp_maze_layout = stdc(file("optimisation.lp", "r"),
                               ASP_BASEDEF,
                               ASP_GF,
                               width=ox, height=oy, 
                               minCliffs=NUM_WALLS-5, maxCliffs=NUM_WALLS+5,
                               minWater=NUM_WATER,
                               maxWater=NUM_WATER,
                               numberOfBases=NUM_PLAYERS,
                               numberOfOilWells=NUM_FREE_OIL,
                               isEmptyMiddle=EMPTY_MIDDLE,
                               isGeographicFeatures=CW_CARE,
                               minCanyonOrCauseway=NUM_CW_MIN,
                               maxCanyonOrCauseway=NUM_CW_MAX)
        for i, solution in enumerate(asp_maze_layout.models()):
            maze = dict(((x, y), 0) for x in xrange(sx) for y in xrange(sy))
            for c, r, z in solution["cellLevel"]:
                if c >= ox or r >= oy:
                    continue
                maze[(c, r)] = z


            BASE_LOCATIONS = [(c, r) for id, c, r in solution["baseLocation"]]
            OIL_LOCATIONS  = [(c, r) for id, c, r in solution["oilLocation"]]

            if len(BASE_LOCATIONS) != NUM_PLAYERS:
                print "Dodgy solution:"
                print solution["raw"]
                print

            show_maze(maze, BASE_LOCATIONS, OIL_LOCATIONS)
            yield StatusUpdate(text="Created maze model %u." % i)

    elif MAPSTYLE == "symmetry":
        assert(ox == oy)
        ASP_BASEDEF = "baseLocation(1, %u, %u).\n" % (1+BASE_EB,
                                                      1+BASE_EB)
        if NUM_PLAYERS == 8:
            ASP_BASEDEF += "baseLocation(2, %u, %u).\n" % (ox-2-BASE_EB,
                                                           1+BASE_EB)
            ASP_BASEDEF += "baseLocation(3, %u, %u).\n" % (1+BASE_EB,
                                                           oy-2-BASE_EB)
        asp_maze_layout = stdc(file("optimisation.lp", "r"),
                               ASP_BASEDEF,
                               ASP_GF,
                               width=ox, height=oy, 
                               minCliffs=NUM_WALLS-5, maxCliffs=NUM_WALLS+5,
                               minWater=NUM_WATER,
                               maxWater=NUM_WATER,
                               numberOfBases=1 if NUM_PLAYERS == 4 else 3,
                               numberOfOilWells=1,
                               isEmptyMiddle=0,
                               isGeographicFeatures=CW_CARE,
                               minCanyonOrCauseway=NUM_CW_MIN,
                               maxCanyonOrCauseway=NUM_CW_MAX)
        for i, solution in enumerate(asp_maze_layout.models()):
            maze = dict(((x, y), 0) for x in xrange(sx) for y in xrange(sy))
            BASE_LOCATIONS = []
            OIL_LOCATIONS = []
            
            for x, y, z in solution["cellLevel"]:
                maze[(x, y)] = z
                maze[(sx-1-x, y)] = z
                maze[(sx-1-x, sy-1-y)] = z
                maze[(x, sy-1-y)] = z

            show_maze(maze, [], [])
            yield StatusUpdate(text="Created maze model %u." % i)


        blot = set()
        BASE_LOCATIONS = [(1+BASE_EB, 1+BASE_EB), (sx-2-BASE_EB, sy-2-BASE_EB)]
        if NUM_PLAYERS >= 4:
            BASE_LOCATIONS.append((sx-2-BASE_EB, 1+BASE_EB))
            BASE_LOCATIONS.append((1+BASE_EB, sy-2-BASE_EB))
        if NUM_PLAYERS >= 8:
            for l in [(sx//2, 1+BASE_EB),
                      (sx//2, sy-2-BASE_EB),
                      (1+BASE_EB, sy//2),
                      (sy-2-BASE_EB, sy//2)]:
                BASE_LOCATIONS.append(l)
                blot |= mapgen.circle(radius=3, base=l)

        blot &= set(maze)

        ACTUAL_WATER = len([x for x in maze if maze[x] == 0])

        ASP_MAPDEF = ""
        for x, y in sorted(maze, key=lambda x: tuple(reversed(x))):
            if (x, y) in blot:
                continue
            ASP_MAPDEF += "cellLevel(%u, %u, %u).\n" % (x, y, maze[(x, y)])
        ASP_BASEDEF = ""
        for id, p in enumerate(sorted(BASE_LOCATIONS,
                                      key=lambda x: tuple(reversed(x)))):
            ASP_BASEDEF += "baseLocation(%u, %u, %u).\n" % (id+1, p[0], p[1])

        asp_oil_layout = stdc(file("optimisation.lp", "r"),
                              ASP_MAPDEF,
                              ASP_BASEDEF,
                              width=sx, height=sy,
                              minCliffs=(NUM_WALLS_SYM-17
                                         if len(blot)
                                         else 0),
                              maxCliffs=(NUM_WALLS_SYM+17
                                         if len(blot)
                                         else sx*sy),
                              minWater=min(ACTUAL_WATER, NUM_WATER_SYM),
                              maxWater=max(ACTUAL_WATER, NUM_WATER_SYM),
                              numberOfBases=NUM_PLAYERS,
                              numberOfOilWells=NUM_FREE_OIL,
                              isEmptyMiddle=0,
                              isGeographicFeatures=0,
                              minCanyonOrCauseway=0,
                              maxCanyonOrCauseway=0
                              )
        OIL_LOCATIONS = None
        for i, oilsolution in enumerate(asp_oil_layout.models()):
            OIL_LOCATIONS = []
            for x, y, z in oilsolution["cellLevel"]:
                maze[(x, y)] = z
            for id, x, y in oilsolution["oilLocation"]:
                OIL_LOCATIONS.append((x, y))
            print "-"*80
            show_maze(maze, BASE_LOCATIONS, OIL_LOCATIONS)
            print OIL_LOCATIONS
            print "-"*80
            yield StatusUpdate(text="Created oil/middle players model %u." % i)

        if OIL_LOCATIONS == None:
            print solution["raw"]
            raise ValueError, "Could not compute oil models."
        

    # Recompute cliffs
    solution["cliffRight"] = []
    solution["cliffDown"] = []
    for y in xrange(sy):
        for x in xrange(sx):
            if x < sx-1 and abs(maze[(x, y)] - maze[(x+1, y)]) >= 2:
                OK = True
                if (maze[(x, y)] == 0 or maze[(x+1, y)] == 0):
                    if abs(maze[(x, y)] - maze[(x+1, y)]) == 2:
                        OK = False
                if OK:
                    solution["cliffRight"].append((x, y))
            if y < sy-1 and abs(maze[(x, y)] - maze[(x, y+1)]) >= 2:
                OK = True
                if (maze[(x, y)] == 0 or maze[(x, y+1)] == 0):
                    if abs(maze[(x, y)] - maze[(x, y+1)]) == 2:
                        OK = False
                if OK:
                    solution["cliffDown"].append((x, y))

    show_maze(maze, BASE_LOCATIONS, OIL_LOCATIONS)

    print "Bases:", BASE_LOCATIONS
    print "Oil wells:", OIL_LOCATIONS

    yield StatusUpdate(text="Created final map model.")

    #--------------------------------------------------------------------------
    # Abstract map grid
    #--------------------------------------------------------------------------

    JITTER_SCALE = float(SCALE)-3.0

    grid = {}
    grid_cells = {}
    edge_right = {}
    edge_down = {}
    attempt = 0
    last_remain = None

    while True:
        # Define grid for yet undefined points
        for p in ((x, y) for x in xrange(sx+1) for y in xrange(sy+1)):
            if p in grid:
                continue
            
            grid[p] = map(float, mapgen.vec_scale(p, SCALE))
            for i in xrange(2):
                jitter = JITTER_SCALE * (random.random() - 0.5)
                for bl in [tmp
                           for tmp in BASE_LOCATIONS
                           if p in mapgen.need_raising([tmp])]:
                    if p[i] == bl[i]:
                        jitter = -JITTER_SCALE*random.uniform(0.45, 0.5)
                    elif p[i] == bl[i]+1:
                        jitter =  JITTER_SCALE*random.uniform(0.45, 0.5)

                if p[i] > 0 and p[i] < (sx if i == 0 else sy):
                    grid[p][i] += jitter
                if p[i] == (sx if i == 0 else sy):
                    grid[p][i] -= 1.0

            grid[p] = tuple(map(lambda x: int(round(x)), grid[p]))

        # Define edges for yet undefined edges
        for p in ((x, y) for x in xrange(sx) for y in xrange(sy)):
            if p not in edge_right:
                if p[0] < sx-1:
                    edge_right[p] = draw_edge(grid[(p[0]+1, p[1])],
                                              grid[(p[0]+1, p[1]+1)],
                                              False)
            if p not in edge_down:
                if p[1] < sy-1:
                    edge_down[p] = draw_edge(grid[(p[0], p[1]+1)],
                                             grid[(p[0]+1, p[1]+1)],
                                             True)

        debug.showset(reduce(lambda x, y: x | y,
                             chain(edge_right.itervalues(),
                                   edge_down.itervalues())))

        # Draw cells for yet undrawn cells
        missing_tiles = set(tp.valid_tiles)
        for (x, y) in maze:
            ol = set()
            ol |= set(edge_right.get((x-1, y), []))
            ol |= set(edge_right.get((x, y), []))
            ol |= set(edge_down.get((x, y-1), []))
            ol |= set(edge_down.get((x, y), []))

            if (x, y) not in grid_cells:
                grid_cells[(x, y)] = flood_fill(ol, ((x*SCALE + SCALE//2),
                                                     (y*SCALE + SCALE//2)))

            missing_tiles -= grid_cells[(x, y)]
            missing_tiles -= ol

        # This is clumsy as anything, but it avoid random missing tiles
        if len(missing_tiles) == 0:
            break
        else:
            # Do it one at a time
            for mt in missing_tiles:
                nb = set(mapgen.axis(n=2, base=mt))
                nb_tiles = filter(lambda x: len(grid_cells[x] & nb),
                                  grid_cells)

                for x, y in mapgen.need_raising(set(nb_tiles)):
                    # Clear intersection
                    if (x, y) in grid:
                        del grid[(x, y)]
                    # Clear the 4 cliffs attached to it
                    for p in [(x-1, y-1), (x-1, y)]:
                        if p in edge_right:
                            del edge_right[p]
                    for p in [(x-1, y-1), (x, y-1)]:
                        if p in edge_down:
                            del edge_down[p]
                    # Clear the 4 grid cells which were defined by the above
                    for p in [(x-1, y-1), (x, y-1), (x-1, y), (x, y)]:
                        if p in grid_cells:
                            del grid_cells[p]

        # Reset attempt count if we did better than last time
        if not last_remain:
            last_attempt = len(missing_tiles)
        if len(missing_tiles) >= last_remain:
            attempt += 1
        else:
            last_remain = len(missing_tiles)
            attempt = 0

        # Completely trash scenarios that blatantly won't get a resolution
        if attempt >= 30:
            last_attempt = None
            attempt = 0
            grid.clear()
            grid_cells.clear()
            edge_down.clear()
            edge_right.clear()


        yield StatusUpdate(text=("Creating terrain grid (%u left to adjust, "
                                 "attempt %u)." %
                                 (len(missing_tiles), attempt)))

    #--------------------------------------------------------------------------
    # Draw the actual cliffs and define height map 
    #--------------------------------------------------------------------------

    # Define cliffs
    for (x, y) in maze:
        if (x, y) in solution["cliffRight"]:
            tp.cliff_tiles |= set(edge_right[(x, y)])
        if (x, y) in solution["cliffDown"]:
            tp.cliff_tiles |= set(edge_down[(x, y)])
    debug.showset(tp.cliff_tiles)

    # Draw cliffs
    cliff_tailends = set()
    for ct in sorted(tp.cliff_tiles):
        nb = set(mapgen.axis(n=1, base=ct)) & tp.cliff_tiles
        # Flange edge tiles
        if ct[0] == 0:
            nb.add(mapgen.vec_sub(ct, (1, 0)))
        if ct[0] == sx - 1:
            nb.add(mapgen.vec_add(ct, (1, 0)))
        if ct[1] == 0:
            nb.add(mapgen.vec_sub(ct, (0, 1)))
        if ct[1] == sy - 1:
            nb.add(mapgen.vec_add(ct, (0, 1)))

        rnb = set(map(lambda x: mapgen.vec_sub(x, ct), nb))
        mapgen.deltex(tp.hmap[ct])

        lo = [0, 0,
              0, 0]
        if (0, 1) in rnb:
            lo[0] = 1
            lo[1] = 1
        if (0, -1) in rnb:
            lo[2] = 1
            lo[3] = 1
        if (1, 0) in rnb:
            lo[0] = 1
            lo[2] = 1
        if (-1, 0) in rnb:
            lo[1] = 1
            lo[3] = 1

        if len(rnb) == 1:
            tmp = filter(lambda x: x[1] == 0, enumerate(lo))
            lo[tmp[random.randint(0, 1)][0]] = 1
            tp.hmap[ct].update(tp.ts.gettrans(tp.deftex,
                                              "Cliff",
                                              tuple(lo)))
            cliff_tailends.add(ct)
        elif len(rnb) == 2 or tp.ts.cliff_x is None:
            if (-1, 0) in rnb and (1, 0) in rnb:
                tp.hmap[ct].update(tp.ts.gettrans(tp.deftex,
                                                  "Cliff",
                                                  (1, 1, 0, 0)))
            elif (0, -1) in rnb and (0, 1) in rnb:
                tp.hmap[ct].update(tp.ts.gettrans(tp.deftex,
                                                  "Cliff",
                                                  (0, 1, 0, 1)))
            else:
                tp.hmap[ct].update(tp.ts.gettrans(tp.deftex,
                                                  "Cliff",
                                                  tuple(lo)))
        else:
            tp.hmap[ct]["tex"] = tp.ts.cliff_x

    # Set up initial height map
    for p in maze:
        h = int((255.0/(DEPTH-2)) * (maze[p]-1))
        for t in mapgen.need_raising(grid_cells[p]) & tp.valid_tiles:
            tp.hmap[t]["height"] = h

    # Terrain smoothing
    tp.reachable_tiles -= tp.cliff_tiles
    tp.paintable_tiles -= tp.cliff_tiles | mapgen.border(tp.cliff_tiles)
    tp.nonfinal_tiles -= tp.cliff_tiles

    tp.blur_terrain(radius=1, exclude=(tp.valid_tiles -
                                       mapgen.need_raising(tp.cliff_tiles)),
                    weight=4)
    for i in xrange(12-DEPTH):
        tp.blur_terrain(radius=1,
                        exclude=mapgen.need_raising(tp.cliff_tiles -
                                                    cliff_tailends),
                        weight=2)
    tp.blur_terrain(radius=max(3, 9-DEPTH), exclude=set(), weight=60)

    # Jitter cliffs
    for t in mapgen.need_raising(tp.cliff_tiles) & tp.valid_tiles:
        tp.hmap[t]["height"] += random.randint(-10, 10)
    
    yield StatusUpdate(text="Created cliffs.")

    #--------------------------------------------------------------------------
    # Lakes
    #--------------------------------------------------------------------------

    grid_lakes = set(filter(lambda x: maze[x] == 0, maze))
    laketiles = set()
    for gl in grid_lakes:
        tiles = grid_cells[gl]
        tiles &= tp.paintable_tiles
        tiles |= mapgen.randomborder(tiles, 25)
        tiles &= tp.paintable_tiles
        tiles |= mapgen.randomborder(tiles, 5)
        tiles &= tp.paintable_tiles
        laketiles |= tiles

    if len(laketiles):
        tp.add_water(laketiles)

        # Make sure lakes don't look stupid
        for t in tp.terrains[-1]["tiles"] | tp.terrains[-1]["border"]:
            if t in tp.valid_tiles:
                tp.hmap[t]["height"] = 0

    yield StatusUpdate(text="Created lakes.")

    #--------------------------------------------------------------------------
    # Roads for arizona and rockies tileset (and reachability for urban)
    #--------------------------------------------------------------------------

    # Choose points to draw roads to
    # Only the bare essentials on urban
    points = set()
    maze_points = set(BASE_LOCATIONS)
    if tp.ts.name in ("Arizona", "Rockies"):
        maze_points |= set(OIL_LOCATIONS)
        for i in xrange((sx*sy)//25):
            maze_points.add((random.randint(0, sx-1),
                             random.randint(0, sy-1)))

    print "Going to visit the following points:", maze_points
    for bl in maze_points:
        # Avoid lakes
        if maze[bl] == 0:
            continue
        tmp = set(grid_cells[bl])
        tmp &= tp.paintable_tiles
        if len(tmp):
            points.add(random.choice(list(tmp)))

    # List of edges (roads) to draw. Ensure that there is a route from any
    # of the interesting points to any other, and then also add a couple
    # of extra edges.
    connected_points = set([points.pop()])
    edges = set()
    while len(points):
        a = random.choice(list(connected_points))
        b = random.choice(list(points))
        points.remove(b)
        connected_points.add(b)
        edges.add(tuple(sorted([a, b])))

    # Extra roads for non-urbans maps:
    if tp.ts.name in ("Arizona", "Rockies"):
        for i in xrange(SCALE//2):
            a = random.choice(list(connected_points))
            b = random.choice(list(connected_points - set([a])))
            edges.add(tuple(sorted([a, b])))

    # Actually draw a road for each edge
    VT = tp.paintable_tiles.copy()
    the_road = set()
    for i, (a, b) in enumerate(edges):
        r = route(a, b, VT, the_road)
        print "Computed route from", a, "->", b,
        print "(%u tiles)" %len(r) if r else "(no route)"
        if r == None:
            continue
        the_road |= frozenset(r)
        yield StatusUpdate(text="Created path %u/%u." % (i+1, len(edges)))

    # This is enough for reachability. The rest is about drawing the road
    if tp.ts.name in ("Arizona", "Rockies"):
        base_tex = tp.ts.simple_texpath(tp.deftex, "Track")
        base_tex = random.choice(list(base_tex)) if base_tex else tp.deftex

        # Check for X-Roads
        try:
            tp.ts.gettrans(base_tex, "Track", (1, 1, 1, 1))
            XROADS = True
        except KeyError:
            XROADS = False

        # If the tileset doesn't have X-roads, then remove any instances
        # of them.
        if not XROADS:
            xtiles = frozenset(rt
                               for rt in the_road
                               if (mapgen.axis(n=1, base=rt) &
                                   tp.valid_tiles) <= the_road)

            print "Found %u x-roads in %u tiles." % (len(xtiles), len(the_road))
            debug.showset(the_road)
            for xt in xtiles:
                if xt not in the_road:
                    continue
                if not ((mapgen.axis(n=1, base=xt) & tp.valid_tiles)
                        <= the_road):
                    print "Skipping", xt
                    continue
                t = random.choice(list(mapgen.axis(n=1, base=xt) &
                                       tp.valid_tiles) + [xt])
                the_road.remove(t)
                print "Deleted road tile %s." % str(t)

        # For tilesets (rockies) where roads can't be drawn on the default
        # tiles, draw an extra texture feature.
        if base_tex != tp.deftex:
            print "Creating %s road helper feature." % base_tex
            f = set(the_road)
            TEX_SIZE = random.randint(5, 15)
            for x in xrange(TEX_SIZE):
                f |= mapgen.randomborder(f, max(TEX_SIZE - x, 5))
                f &= tp.paintable_tiles
            tp.add_texfeature(f, base_tex)

        # Make sure that the section below doesn't scribble over the road
        tp.paintable_tiles -= the_road | mapgen.border(the_road)
        tp.nonfinal_tiles -= the_road

    #--------------------------------------------------------------------------
    # Random texture features
    #--------------------------------------------------------------------------
    
    if WZ_VERSION == "trunk" and tp.ts.name == "Rockies":
        # We've got a different algorithm for that
        NUM_TEXF = 0
    else:
        NUM_TEXF = int((sx*sy*SCALE) // 10)

    for n in xrange(NUM_TEXF):
        # This can happen for water filled maps
        if len(tp.paintable_tiles) == 0:
            break

        # Pick a random point, and then grow
        f = set([mapgen.choose_one(tp.paintable_tiles)])
        if n <= NUM_TEXF // 3:
            TEX_SIZE = random.randint(35, 45)
        else:
            TEX_SIZE = random.randint(5, 15)
        for x in xrange(TEX_SIZE):
            f |= mapgen.randomborder(f, max(TEX_SIZE - x, 10))
            f &= tp.paintable_tiles

        tp.add_texfeature(f)

        yield StatusUpdate(text="Created texture patch %u/%u." %
                           (n, NUM_TEXF))
        

    yield StatusUpdate(text="Created texture features.")

    #--------------------------------------------------------------------------
    # Render terrain
    #--------------------------------------------------------------------------

    tp.render()
    
    # Roads have to be painted after the actual terrain
    if tp.ts.name in ("Arizona", "Rockies"):
        tp.paint_road([the_road], name="Track", urban=False)

    yield StatusUpdate(text="Rendered terrain.")

    #--------------------------------------------------------------------------
    # Roads for urban tileset
    #--------------------------------------------------------------------------
    
    if tp.ts.name == "Urban":
        # Tiles on which a road can begin and end
        VALID_TRANSTILES = frozenset(tp.ts.get_alltrans("Road"))

        def make_road(fixed_coord, fc_value, joinables=set(), ignore=set()):
            other_coord = int(not fixed_coord)
            mapsize = (sx*SCALE, sy*SCALE)
            orig_fc_value = fc_value

            cr = []
            def do_trim():
                for end in [0, -1]:
                    for i in xrange(len(cr)):
                        if cr[end] in joinables:
                            break
                        if (cr[end] not in tp.nonfinal_tiles or
                            cr[end] in ignore or
                            (tp.ts.getname(tp.hmap[cr[end]]["tex"]) not in
                             VALID_TRANSTILES)):
                            del cr[end]
                        else:
                            break
                        if len(cr) == 0:
                            break

            for i in xrange(mapsize[other_coord]):
                tmp = [0, 0]
                tmp[fixed_coord] = fc_value
                tmp[other_coord] = i
                t = tuple(tmp)

                if (t not in tp.reachable_tiles or \
                        t in ignore or \
                        t in joinables) and len(cr):

                    if t in joinables:
                        cr.append(t)

                    if len(cr) >= 3:
                        do_trim()
                    if len(cr) >= 3:
                        yield cr
                    cr = []

                    if t in joinables:
                        if fc_value == orig_fc_value:
                            fc_value += -1 + 2*random.randint(0, 1)
                        else:
                            fc_value = orig_fc_value
                        tmp[fixed_coord] = fc_value
                        tmp[other_coord] = i
                        cr.append(tuple(tmp))
                else:
                    cr.append(t)
            if len(cr) >= 3:
                do_trim()
            if len(cr) >= 3:
                yield cr
        # end of make_road

        joinables = set()
        invalid = set()

        roads = []

        print "Creating horizontal roads..."
        for y in (y*SCALE + SCALE//2 + random.randint(-1, 1)
                  for y in xrange(sy-1)):
            tmp = [y]
            if random.randint(0, 10) >= 6:
                tmp.append(y -1 + 2*random.randint(0, 1))

            for yy in tmp:
                for r in make_road(1, yy):
                    roads.append(frozenset(r))
                    joinables |= set(r[1:-1])
                    invalid |= set(mapgen.axis(n=1, base=r[0])) - joinables
                    invalid |= set(mapgen.axis(n=1, base=r[-1])) - joinables
                    invalid |= set([r[0], r[-1]])

        print "Creating vertical roads..."
        for x in (x*SCALE + SCALE//2 + random.randint(-1, 1)
                  for x in xrange(sx-1)):
            for r in make_road(0, x,
                               joinables=joinables,
                               ignore=invalid):
                roads.append(frozenset(r))

        tp.paint_road(roads)
 
        yield StatusUpdate(text="Created urban roads.")
    
    #--------------------------------------------------------------------------
    # Bases and other oil wells
    #--------------------------------------------------------------------------

    if ASP_FOR_BASES:
        # Set up some stuff needed for the nifty (but slow) ASP based bases
        LP_BASES = file("bases.lp", "r").read()
        ASP_BASES_INPUT = "".join("building(%s).\nsize(%s,%u).\n" % \
                                      (x[0], x[0], x[1])
                                  for x in ([("hq", 2),
                                             ("power", 2)] +
                                            [("truck%u" % x, 1)
                                             for x in xrange(2)] +
                                            [("research%u" % x, 2)
                                             for x in xrange(2)] +
                                            [("factory%u" % x, 3)
                                             for x in xrange(2)] +
                                            [("oil%u" % x, 1)
                                             for x in xrange(NUM_PLAYER_OIL)]))

        stdc = partial(asp.compile,
                       LP_BASES,
                       ASP_BASES_INPUT)

    for pid, bl in enumerate(BASE_LOCATIONS):
        # Define base space and then grow a little bit if possible
        initial_tiles = grid_cells[bl]
        for x in xrange(4):
            initial_tiles |= mapgen.border(initial_tiles)
            initial_tiles &= tp.reachable_tiles
            initial_tiles &= tiles_in_border
            initial_tiles -= set(t
                                 for t in initial_tiles
                                 if len(mapgen.field(1, t) &
                                        initial_tiles) <= 4)
            if len(initial_tiles) > 280:
                break
        tp.paintable_tiles -= initial_tiles

        # Place wall on border, and place buildings leaving at least 1 tile gap
        base_border = (mapgen.border(tp.valid_tiles - initial_tiles)
                       & grid_cells[bl])
        base_tiles = initial_tiles - base_border
        base_border = mapgen.border(base_tiles)
        base_tiles &= grid_cells[bl]

        # For small scales we won't have enough space otherwise
        if (not ASP_FOR_BASES or
            len(base_tiles) >= (90 # Reasonably tight packing of main buildings
                                + NUM_PLAYER_OIL
                                + len(mapgen.border(base_border) &
                                      base_tiles))):
            base_tiles -= mapgen.border(base_border)

        # We have a gate on the guaranteed road between each players base
        # For the urban tileset the "invisible reachability" road is used
        debug.showset(base_border, the_road)
        possible_gates = base_border & the_road
        possible_gates = set(filter(lambda x: len(mapgen.axis(n=1, base=x) &
                                                  base_border) >= 2,
                                    possible_gates))
        while len(possible_gates):
            t = mapgen.choose_one(possible_gates)
            base_border -= mapgen.field(n=1, base=t)
            possible_gates -= mapgen.field(n=3, base=t)

        debug.showset(base_border, highlight=base_tiles)

        # Paint walls
        for t in base_border:
            nbr = frozenset(map(operator.itemgetter(0),
                                filter(lambda x: x[1] in base_border,
                                       mapgen.axis_rel(n=1, base=t))))
            if len(nbr) == 2:
                if nbr == set([(-1, 0), (1, 0)]):
                    tp.add_building("A0HardcreteMk1Wall", t, pid)
                elif nbr == set([(0, -1), (0, 1)]):
                    tp.add_building("A0HardcreteMk1Wall", t, pid, dir=90)
                else:
                    tp.add_building("A0HardcreteMk1CWall", t, pid)
            elif len(nbr) == 1:
                tp.add_building(random.choice(["WallTower01",
                                               "WallTower02"]), t, pid)

            else:
                tp.add_building("A0HardcreteMk1CWall", t, pid)
                
        if ASP_FOR_BASES:
            # ASP bases
            tl = (min(p[0] for p in base_tiles), min(p[1] for p in base_tiles))
            br = (max(p[0] for p in base_tiles), max(p[1] for p in base_tiles))
            ASP_UNIQUE_START = "uniqueStart(%u, %u).\n" % \
                tuple(map(int, mapgen.vec_scale(mapgen.vec_add(tl, br), 0.5)))
            ASP_BLOCKED = ""
            for p in mapgen.square(tl[0], tl[1], br[0], br[1]):
                if p not in base_tiles:
                    ASP_BLOCKED += "block(%u, %u).\n" % p

            for i, bsol in enumerate(stdc(ASP_UNIQUE_START,
                                          ASP_BLOCKED,
                                          xmin=tl[0], xmax=br[0],
                                          ymin=tl[1], ymax=br[1]).models()):
                buildings = {}
                for n, x in bsol["xOrigin"]:
                    if not n.startswith("oil") and not n.startswith("truck"):
                        x += 1
                    buildings[n] = x
                for n, y in bsol["yOrigin"]:
                    if not n.startswith("oil") and not n.startswith("truck"):
                        y += 1
                    buildings[n] = (buildings[n], y)

                yield StatusUpdate(text="Computing player base %u/%u"
                                   " (model %u)." % (pid+1, NUM_PLAYERS, i+1))
        else:
            # Quick and dirty bases
            buildings = {
                #"factory1": pick_first(base_tiles, size=3),
                #"factory2": pick_first(base_tiles, size=3),
                "power": pick_first(base_tiles, size=2),
                "hq": pick_first(base_tiles, size=2),
                #"research1": pick_first(base_tiles, size=2),
                #"research2": pick_first(base_tiles, size=2),
                "truck1": pick_first(base_tiles, size=1),
                "truck2": pick_first(base_tiles, size=1)
                }
            for i in xrange(NUM_PLAYER_OIL):
                buildings["oil%u" % i] = base_tiles.pop()

        for bn, p in buildings.iteritems():
            if bn.startswith("power"):
                tp.add_building("A0PowerGenerator", p, pid)
            elif bn.startswith("hq"):
                tp.add_building("A0CommandCentre", p, pid)
            elif bn.startswith("factory"):
                tp.add_building("A0LightFactory", p, pid)
            elif bn.startswith("research"):
                tp.add_building("A0ResearchFacility", p, pid)
            elif bn.startswith("oil"):
                tp.add_feature("OilResource", p)
                tp.add_building("A0ResourceExtractor", p, pid)
            elif bn.startswith("truck"):
                tp.add_droid({
                        "name" : "ConstructorDroid",
                        "pos": p,
                        "player": pid
                        })
            else:
                raise ValueError, "Could not deal with item '%s'." % bn
        
    yield StatusUpdate(text="Created player bases.")

    #--------------------------------------------------------------------------
    # Free oil wells
    #--------------------------------------------------------------------------

    for bp in OIL_LOCATIONS:
        base_tiles = (grid_cells[bp] & tp.reachable_tiles)
        base_tiles &= tiles_in_border
        if len(base_tiles) == 0:
            p = mapgen.vec_add(mapgen.vec_scale(bp, SCALE),
                               (SCALE//2, SCALE//2))
            base_tiles = mapgen.circle(radius=5, base=p)

        while len(base_tiles) and len(base_tiles) < FREE_OIL_CLUSTER_SIZE * 4:
            base_tiles |= mapgen.border(base_tiles) & tp.reachable_tiles
            base_tiles &= tiles_in_border

        if len(base_tiles) < FREE_OIL_CLUSTER_SIZE:
            debug.showset(tp.reachable_tiles, highlight=grid_cells[bp])
            raise ValueError, "Could not place oil well cluster @ %s" % repr(bp)

        # We need this to fall back on as pick_random removes more than it
        # can chew
        base_tiles_orig = base_tiles.copy()

        for i in xrange(FREE_OIL_CLUSTER_SIZE):
            try:
                well_pos = pick_random(base_tiles)
                base_tiles_orig.remove(well_pos)
            except IndexError:
                well_pos = base_tiles_orig.pop()
            tp.add_feature("OilResource", well_pos)

    yield StatusUpdate(text="Created oil well clusters.")

    #--------------------------------------------------------------------------
    # Other generic shit
    #--------------------------------------------------------------------------
    
    # All potential tiles that can be used for a feature
    valid_feature_tiles = set(tp.reachable_tiles)
    valid_feature_tiles &= tiles_in_border
    valid_feature_tiles -= reduce(lambda x, y: x | y,
                                  (grid_cells[x]
                                   for x in chain(BASE_LOCATIONS,
                                                  OIL_LOCATIONS)),
                                  set())
    
    def place_features(features, tiles, amount, VFT=valid_feature_tiles):
        ALL_SIZES = set(f["size"] for f in features.itervalues())
        possibles = {}
        for s in ALL_SIZES:
            possibles[s] = set(x for x in tiles
                               if mapgen.building(s[0], s[1], base=x) <= \
                                   tiles)

        for i in xrange(1, amount+1):
            fn = mapgen.choose_one(features)
            size = features[fn]["size"]
            print "Placing feature %u/%u." % (i, amount)
            
            # Choose location if one exists
            if len(possibles[size]) == 0:
                continue
            else:
                pos = mapgen.choose_one(possibles[size])

            # Adujust possible locations for other features
            for s in ALL_SIZES:
                space = mapgen.building(size[0], size[1], base=pos)
                space |= mapgen.border(space)
                for coord in xrange(2):
                    def foo(amount):
                        tmp = [0, 0]
                        tmp[coord] = amount
                        return tuple(tmp)

                    for j in xrange(s[coord]-1):
                        space |= set(mapgen.vec_add(x, foo(1)) for x in space)
                    for j in xrange(max(s[coord]-2, 0)):
                        space |= set(mapgen.vec_add(x, foo(-1)) for x in space)
                possibles[s] -= space

            # Make sure the ground is even
            tmp = mapgen.need_raising(mapgen.building(size[0], size[1],
                                                      base=pos))
            h = sum(tp.hmap[x]["height"] for x in tmp) // len(tmp)
            for t in tmp:
                tp.hmap[t]["height"] = h

            # Place feature on map
            tp.add_feature(fn, pos, width=size[0], height=size[1])

        # Update valid feature tiles for convenience
        VFT &= tp.reachable_tiles
    
    def near_road(dst):
        rv = set(tp.road_tiles)
        for i in xrange(dst):
            rv |= mapgen.border(rv)
            rv &= valid_feature_tiles
        rv -= tp.road_tiles
        return rv

    if tp.ts.name == "Arizona":
        TILES_MISC = ((tp.ts.get_set("Red Dirt") |
                       tp.ts.get_set("Yellow Dirt") |
                       tp.ts.get_set("Yellow Dirt Shrubs") |
                       tp.ts.get_set("Red Clay")) -
                      (tp.ts.get_set("Tiles") |
                       tp.ts.get_set("Dark Clay") |
                       tp.ts.get_set("Water")))
        TILES_ROCKS = set(x[0] for x in tp.ts.terraintiles["Red Dirt"])
        TILES_TECH = set(x[0] for x in tp.ts.terraintiles["Tiles"])

        # Rocks
        print "Rocks..."
        vt_rocks = set(filter(lambda t: tp.hmap[t]["tex"] in TILES_ROCKS,
                              valid_feature_tiles))
        vt_rocks -= tp.road_tiles
        
        place_features(tp.ts.feat_rocks, vt_rocks, len(vt_rocks) // 40)
        
        # Barbarian towns
        for i in xrange(int(math.sqrt(sx*sy) // 4)):
            print "Barbarian town %u..." % i
            vt_misc = set(filter(lambda t: tp.hmap[t]["tex"] in TILES_MISC,
                                 valid_feature_tiles))
            vt_town = near_road(1)
            for et in chain(edge_right.itervalues(), edge_down.itervalues()):
                vt_town -= et
            
            # This can happen on water heavy maps
            if len(vt_town) == 0:
                continue
            town_pos = mapgen.choose_one(vt_town)

            tmp = filter(lambda x: town_pos in grid_cells[x], grid_cells)
            print town_pos, tmp

            if len(tmp):
                town_tiles = set(grid_cells[tmp[0]])
            else:
                print "Skipping town %u." % i
                continue
                    
            town_tiles &= valid_feature_tiles
            n = 0
            while len(town_tiles) < 100 and n <= 4:
                town_tiles |= mapgen.randomborder(town_tiles, 40) & vt_misc
                town_tiles &= valid_feature_tiles
                n += 1
            town_tiles -= tp.road_tiles
            town_tiles &= vt_misc

            print "Tiles:", len(town_tiles)
            place_features(tp.ts.feat_buildings, town_tiles,
                           len(town_tiles) // 8)

            town_tiles &= valid_feature_tiles
            place_features(tp.ts.feat_cars, town_tiles,
                           len(town_tiles) // 12)

        # Tech stuff
        print "Tech buildings..."
        vt_tech = set(filter(lambda t: tp.hmap[t]["tex"] in TILES_TECH,
                             valid_feature_tiles))
        place_features(tp.ts.feat_techbuildings, vt_tech,
                       len(vt_tech) // 12)

    elif tp.ts.name == "Urban":
        # Buildings and stuff. Don't place buildings on the highest tiles
        valid_building_tiles = near_road(3)
        valid_building_tiles -= \
            reduce(lambda x, y: x | y,
                   (grid_cells[x]
                    for x in filter(lambda p: maze[p] in (DEPTH-1,),
                                    maze)),
                   set())

        place_features(tp.ts.feat_buildings, valid_building_tiles,
                       int((sx*sy) // 2))

        # Cars and stuff
        valid_road_tiles = set(tp.road_tiles)
        valid_road_tiles |= mapgen.border(valid_road_tiles)
        valid_road_tiles &= valid_feature_tiles
        valid_road_tiles &= tp.reachable_tiles

        place_features(tp.ts.feat_road, valid_road_tiles,
                       len(valid_road_tiles) // 30)

    elif tp.ts.name == "Rockies":
        # Buildings
        place_features(tp.ts.feat_buildings, near_road(1),
                       len(tp.road_tiles) // 20)

        # Trees...
        TILES_SNOW = (tp.ts.get_set("Dirt Snow") |
                      tp.ts.get_set("Grass Snow") |
                      tp.ts.get_set("Snow"))

        # ...not on snow
        valid_tree_tiles = \
            set(filter(lambda t: tp.hmap[t]["tex"] not in TILES_SNOW,
                       valid_feature_tiles))
        valid_tree_tiles -= tp.road_tiles
        feat_trees = dict(filter(lambda x: not x[0].startswith("TreeSnow"),
                                 tp.ts.feat_misc.iteritems()))
        place_features(feat_trees, valid_tree_tiles,
                       len(valid_tree_tiles) // 75)


        # ...on snow
        valid_snowtree_tiles = \
            set(filter(lambda t: tp.hmap[t]["tex"] in TILES_SNOW,
                       valid_feature_tiles))
        valid_tree_tiles -= tp.road_tiles
        feat_snowtrees = dict(filter(lambda x: x[0].startswith("TreeSnow"),
                                     tp.ts.feat_misc.iteritems()))
        place_features(feat_snowtrees, valid_snowtree_tiles,
                       len(valid_snowtree_tiles) // 75)


    yield StatusUpdate(text="Created misc features.") 
   
    yield tp
