#------------------------------------------------------------------------------
#    Copyright (C) 2009 Florian Schanda <florian@schanda.org.uk>
#
#    This file is part of the Warzone Map Tools.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program. If not, see <http://www.gnu.org/licenses/>.
#------------------------------------------------------------------------------

from mapwrite import install_map

from itertools import chain
import operator
import math
import random

import blur

import debug

#
# 2D Vector functions
#

def vec_add(a, b):
    return (a[0]+b[0], a[1]+b[1])

def vec_sub(a, b):
    return (a[0]-b[0], a[1]-b[1])

def vec_mul(a, b):
    return (a[0]*b[0], a[1]*b[1])

def vec_scale(a, scalar):
    return (a[0]*scalar, a[1]*scalar)

def vec_len(a):
    return math.sqrt(a[0]*a[0] + a[1]*a[1])

def vec_normal2d(a):
    l = math.sqrt(a[0]*a[0] + a[1]*a[1])
    return (a[1]/l, -a[0]/l)

#
# Generic functions
#

def circle(base, radius):
    return set((base[0]+x, base[1]+y)
               for x in xrange(-radius, radius+1)
               for y in xrange(-radius, radius+1)
               if x*x + y*y <= radius*radius)

def field_rel(n, base=None):
    if base is None:
        base = (0, 0)
    return set(((x, y), (base[0]+x, base[1]+y))
               for x in xrange(-n, n+1)
               for y in xrange(-n, n+1))

def field(n, base=None):
    if base is None:
        base = (0, 0)
    return set((base[0]+x, base[1]+y)
               for x in xrange(-n, n+1)
               for y in xrange(-n, n+1))

def axis_rel(n, base=None):
    if base is None:
        base = (0, 0)
    return set(((i if c == 0 else 0, i if c == 1 else 0),
                (base[0] + (i if c == 0 else 0),
                 base[1] + (i if c == 1 else 0)))
               for i in xrange(-n, n+1)
               for c in xrange(2)
               if i != 0)

def axis(n, base=None):
    if base is None:
        base = (0, 0)
    return set((base[0] + (i if c == 0 else 0),
                base[1] + (i if c == 1 else 0))
               for i in xrange(-n, n+1)
               for c in xrange(2)
               if i != 0)

def square(x1, y1, x2, y2):
    return set((x, y)
               for x in xrange(x1, x2+1)
               for y in xrange(y1, y2+1))

def building(width, height, base=None):
    FOO = ((0, 1), (-1, 1), (-1, 2))
    if base is None:
        base = (0, 0)
    return set(vec_add((x, y), base)
               for x in xrange(*FOO[width-1])
               for y in xrange(*FOO[height-1]))

def valid_tiles(t, mapw, maph):
    return set(filter(lambda x: (x[0] >= 0 and x[1] >= 0 and
                                 x[0] < mapw and x[1] < maph),
                      t))

def deltex(tile):
    for tmp in ("rot", "flipx", "flipy"):
        if tmp in tile:
            del tile[tmp]
    tile["tex"] = 22

def choose_one(x):
    return random.choice(list(x))

def clamp(x, xmin, xmax):
    return min(max(xmin, x), xmax)

def diagonals(w, h, down=True):
    """ Iterate over a grid, yielding downwards diagonals. """
    def is_valid(p):
        return p[0] >= 0 and p[1] >= 0 and p[0] < w and p[1] < h
    
    p = (0, h-1) if down else (0, 0)
    sdir = (1, 0)
    mdir = (1, 1) if down else (1, -1)
    yield [p]
    if (w+h)-2 == 0:
        return
    for n in xrange((w+h)-2):
        # Shift
        p = vec_add(p, sdir)

        # Set up new directions
        if sdir == (1, 0):
            sdir = (0, -1) if down else (0, 1)
        else:
            sdir = (1, 0)
        mdir = tuple(-i for i in mdir)

        # Move along
        rv = []
        if is_valid(p):
            rv.append(p)
        while True:
            tmp = vec_add(p, mdir)
            if is_valid(tmp):
                p = tmp
            else:
                break
            rv.append(p)
        yield rv

#
# Feature manipulation
#

def border(base):
    return reduce(operator.ior, (field(1, base=t) for t in base), set()) - base

def randomborder(base, chance):
    return frozenset(filter(lambda x: random.randint(1, 100) <= chance,
                            border(base)))


PUSHMAP = {
    (-1, -1): (0, 0,
               0, 1),
    ( 0, -1): (0, 0,
               1, 1),
    ( 1, -1): (0, 0,
               1, 0),
    (-1,  0): (0, 1,
               0, 1),
    ( 1,  0): (1, 0,
               1, 0),
    (-1,  1): (0, 1,
               0, 0),
    ( 0,  1): (1, 1,
               0, 0),
    ( 1,  1): (1, 0,
               0, 0)
    }

INVALID_TILES = [(0, 1, 1, 0), (1, 0, 0, 1)]

def make_btm(tiles):
    bt = border(tiles)
    rv = {}
    
    # Init
    for t in bt:
        rv[t] = (0, 0, 0, 0)

    # Update border tile map
    for ft in tiles:
        # Watch carefully...
        for relpos, dst in filter(lambda x: x[1] in bt, field_rel(1, base=ft)):
            rv[dst] = tuple(map(lambda x: x[0] | x[1],
                                zip(rv[dst], PUSHMAP[relpos])))
    
    return rv

def develop_feature(feat, nft=None):
    assert(nft)

    btmap = {}

    def merge(p, m):
        btmap[p] = tuple(map(lambda x: x[0] | x[1],
                             zip(btmap.get(p, (0, 0, 0, 0)), m)))

    def putsquare(pos):
        merge((pos[0]  , pos[1]  ), (0, 0,
                                     0, 1))
        merge((pos[0]+1, pos[1]  ), (0, 0,
                                     1, 0))
        merge((pos[0]+1, pos[1]+1), (1, 0,
                                     0, 0))
        merge((pos[0]  , pos[1]+1), (0, 1,
                                     0, 0))

    btmap = make_btm(feat["tiles"])

    problem_tiles = filter(lambda x: btmap[x] in INVALID_TILES, btmap)
    if len(problem_tiles):
        print "Found problem tiles: %s" % problem_tiles
        debug.showset(feat["tiles"], highlight=problem_tiles)

    for i in xrange(3):
        for pt in problem_tiles:
            if btmap[pt] == (1, 0, 0, 1):
                # bottom left or top right
                A = vec_add(pt, (-1, 0)) if vec_add(pt, (-1, 1)) in nft else None
                B = vec_add(pt, (0, -1)) if vec_add(pt, (1, -1)) in nft else None
            else:
                # top left or bottom right
                A = vec_add(pt, (-1, -1)) if vec_add(pt, (-1, -1)) in nft else None
                B = vec_add(pt, (0, 0)) if vec_add(pt, (1, 1)) in nft else None
            if A and B:
                n = random.randint(0, 2)
                n = 1
                if n <= 1:
                    putsquare(A)
                if n >= 1:
                    putsquare(B)
            elif A:
                print "Resolved %s with %s." % (repr(pt), repr(A))
                putsquare(A)
            elif B:
                print "Resolved %s with %s." % (repr(pt), repr(B))
                putsquare(B)

        problem_tiles = filter(lambda x: btmap[x] in INVALID_TILES, btmap)
        if len(problem_tiles) == 0:
            break
        elif i > 0:
            for rpt in problem_tiles:
                feat["tiles"] -= set(field(n=1, base=rpt))
            btmap = make_btm(feat["tiles"])
            problem_tiles = filter(lambda x: btmap[x] in INVALID_TILES, btmap)
        print "After attempt: %u" % i
        debug.showset(feat["tiles"], highlight=problem_tiles)

    if len(problem_tiles):
        debug.showset(feat["tiles"], highlight=problem_tiles)
        raise ValueError, "Despite best attempts, could not draw terrain."

    feat["tiles"] = frozenset(feat["tiles"] |
                              set(x for x in btmap
                                  if btmap[x] == (1, 1, 1, 1)))
    for x in set(btmap) & feat["tiles"]:
        del btmap[x]
    feat["raise"] = need_raising(feat["tiles"])

    feat["btmap"] = btmap
    feat["border"] = frozenset(feat["btmap"].keys())

def used_space(df):
    return frozenset(df["tiles"] | border(df["tiles"]))
    
def need_raising(tiles):
    """ Given a feature which needs to be raised in height, this set
        defines which tiles should be raised.
    """
    return reduce(operator.ior, (set([t,
                                      (t[0]+1, t[1]),
                                      (t[0], t[1]+1),
                                      (t[0]+1, t[1]+1)])
                                 for t in tiles),
                  set())

def need_water_raising(tiles):
    """ Given a water feature which needs to be set in height, this set
        defines which tiles should be raised.
    """
    return reduce(operator.ior, (set([t,
                                      (t[0]+1, t[1]),
                                      (t[0], t[1]+1),
                                      (t[0]+1, t[1]+1)])
                                 for t in tiles | border(tiles)),
                  set())

#
# Feature: Mountains
#

def fracture(line, factor):
    assert(len(line) >= 2)

    rv = []
    rv.append(line[0])

    for segment in xrange(len(line) - 1):
        a, b = line[segment:segment+2]

        mid = vec_scale(vec_add(a, b), 0.5)
        vecdir = vec_sub(b, a)

        n = vec_normal2d(vecdir)
        amount = vec_len(vecdir) * factor
        amount = random.uniform(-amount, amount)

        rv.append(vec_add(mid, vec_scale(n, amount)))

    rv.append(line[-1])

    return rv

def rasterize(line):
    assert(len(line) >= 2)
    RES = 50

    rv = set()
    for segment in xrange(len(line) - 1):
        a, b = line[segment:segment+2]
        dir = vec_sub(b, a)
        ss = 1.0 / (RES * vec_len(dir))
        for i in xrange(int(vec_len(dir))*RES):
            p = vec_add(a, vec_scale(dir, ss*i))
            rv |= set([tuple(map(int, p))])

    return rv


def random_mountain(line, dev=0.3, fractures=3, chance=25, borders=1):
    for x in xrange(fractures):
        line = fracture(line, dev)

    tmp = rasterize(line)
    for x in xrange(borders):
        tmp |= randomborder(tmp, chance)
    return tmp

#
# Generic terrain painter class
#

class TerrainPainter():
    def __init__(self, w, h, ts):
        self.mapw = w
        self.maph = h
        self.ts = ts

        self.deftex = ts.get_default()
        self.def_to_water = list(set(self.ts.get_alltrans(self.deftex)) &
                                 set(self.ts.get_alltrans("Water")))
        if self.deftex in self.ts.get_alltrans("Water"):
            self.def_to_water.append(None)

        print "Default texture:", self.deftex
        print "  to Water via:", self.def_to_water

        self.hmap = {}
        for y in xrange(self.maph):
            for x in xrange(self.mapw):
                self.hmap[(x, y)] = self.ts.getrandom(self.deftex)
                self.hmap[(x, y)]["height"] = random.randint(20, 25)

        self.valid_tiles = frozenset(self.hmap)

        self.mountains = []
        self.waters = []
        self.terrains = []

        self.droids = []
        self.features = []
        self.buildings = []

        # Set of tiles that can be driven to
        self.reachable_tiles = set(self.hmap.iterkeys())

        # Set of tiles that can take a texture feature
        self.paintable_tiles = set(self.hmap.iterkeys())

        # Set of tiles that can take a transition
        self.nonfinal_tiles = set(self.hmap.iterkeys())

        # Other information about terrain
        self.road_tiles = set()
        self.building_tiles = set()
        self.cliff_tiles = set()
        self.water_tiles = set()

    def blur_terrain(self, radius, exclude, weight):
        tmp = dict((p, self.hmap[p]["height"]) for p in self.hmap)
        blur.blur(tmp, radius, exclude, weight)
        for t in self.hmap:
            self.hmap[t]["height"] = tmp[t]
 
    def add_mountain(self, tiles):
        tmp = {
            "tiles": frozenset(tiles),
            }
        develop_feature(tmp, nft=self.nonfinal_tiles)
        
        tmp["cliff"] = valid_tiles(tmp["border"],
                                   self.mapw, self.maph)
        self.mountains.append(tmp)

        self.reachable_tiles -= tmp["tiles"] | tmp["border"]
        self.paintable_tiles -= tmp["border"] | border(tmp["border"])
        self.nonfinal_tiles -= tmp["border"]
        
        print ("Added mountain with %u tiles and %u border tiles." %
               (len(tmp["tiles"]), len(tmp["btmap"])))


    def add_water(self, tiles):
        tmp = {
            "tiles": set(tiles),
            "type": "Water"
            }
        extra_tex = random.choice(self.def_to_water)

        # Standard pruning
        tmp["tiles"] &= self.paintable_tiles

        # Extra transition
        if extra_tex:
            self.add_texfeature(tmp["tiles"], type=extra_tex)
            tmp["tiles"] &= self.paintable_tiles

        develop_feature(tmp, nft=self.nonfinal_tiles)

        self.terrains.append(tmp)

        self.reachable_tiles -= tmp["tiles"] | tmp["border"]
        self.paintable_tiles -= (tmp["tiles"] | tmp["border"] |
                                 border(tmp["border"]))
        self.nonfinal_tiles -= tmp["tiles"] | tmp["border"]
        self.water_tiles |= tmp["tiles"] | tmp["border"]

        print ("Added water with %u tiles and border tiles." %
               len(tmp["tiles"] | tmp["border"]))

    def add_texfeature(self, tiles, type=None):
        tmp = {
            "tiles": frozenset(tiles),
            "type": type
            }
        develop_feature(tmp, nft=self.nonfinal_tiles)

        self.terrains.append(tmp)
        
        self.paintable_tiles -= tmp["border"] | border(tmp["border"])
        self.nonfinal_tiles -= tmp["border"]
        
        print ("Added terrain patch with %u tiles and %u border tiles." %
               (len(tmp["tiles"]), len(tmp["btmap"])))

    def add_droid(self, droid):
        tmp = droid.copy()
        self.reachable_tiles -= set([tmp["pos"]])
        self.droids.append(tmp)

    def add_feature(self, name, position, width=1, height=1):
        tmp = {
            "name": name,
            "pos": position,
            "tpos": ((width % 2) * 64, ((height % 2) * 64))
            }
        if width == 1 and height == 1:
            tmp["dir"] = random.randint(0, 359)
        self.reachable_tiles -= building(width, height, base=position)
        self.building_tiles |= building(width, height, base=position)
        self.features.append(tmp)

    def add_building(self, name, position, player, dir=None):
        SIZE = {
            "A0LightFactory": (3, 3),
            "A0CommandCentre": (2, 2),
            "A0PowerGenerator": (2, 2),
            "A0ResearchFacility": (2, 2),
            "A0ResourceExtractor": (1, 1),
            "A0HardcreteMk1Wall": (1, 1),
            "A0HardcreteMk1CWall": (1, 1),
            "WallTower01": (1, 1),
            "WallTower02": (1, 1),
            }[name]

        tmp = {
            "name": name,
            "pos": position,
            "player": player,
            }
        if dir is not None:
            tmp["dir"] = dir

        self.reachable_tiles -= building(SIZE[0], SIZE[1], base=position)
        self.building_tiles |= building(SIZE[0], SIZE[1], base=position)
        self.buildings.append(tmp)

    def paint_road(self, roads, name="Road", urban=True):
        """ This function will paint an (assumed valid) road on top of the
            current terrain. (This is useful for the urban tileset.)
        """
        road_tiles = set()

        for road in roads:
            print "Painting road with %u tiles." % len(road)
            for rt in road:
                rnb = set(map(lambda x: vec_sub(x, rt),
                              set(axis(n=1, base=rt)) & road))
                # Flange edge tiles
                if rt[0] == 0:
                    rnb.add((-1, 0))
                if rt[0] == self.mapw - 1:
                    rnb.add((1, 0))
                if rt[1] == 0:
                    rnb.add((0, -1))
                if rt[1] == self.maph - 1:
                    rnb.add((0, 1))

                if len(rnb) == 1 and rt in road_tiles:
                    nb = list(rnb)[0]
                    opp = tuple(map(operator.neg, nb))
                    rnb |= set(map(lambda x: vec_sub(x, rt),
                                   set(axis(n=1, base=rt)) & road_tiles))
                    if opp in rnb:
                        rnb.remove(opp)
                    

                lo = map(int, [( 0, -1) in rnb,
                               (-1,  0) in rnb,
                               ( 1,  0) in rnb,
                               ( 0,  1) in rnb])

                if len(rnb) == 1 or not urban:
                    tmp = self.ts.getname(self.hmap[rt]["tex"])
                else:
                    tmp = None

                deltex(self.hmap[rt])
                try:
                    self.hmap[rt].update(self.ts.gettrans(tmp, name, lo))
                except:
                    print tmp, name, lo
                    raise
            road_tiles |= road
        self.road_tiles |= road_tiles
            

    def render(self):
        # Paint mountains
        for mountain in self.mountains:
            # Raise terrain
            nr = valid_tiles(need_raising(mountain["tiles"]),
                             self.mapw, self.maph)
            for t in nr:
                self.hmap[t]["height"] += 120 + random.randint(-5, 15)

            # Raise border terrain
            for t in valid_tiles(border(nr) - mountain["tiles"],
                                 self.mapw, self.maph):
                self.hmap[t]["height"] += 40 + random.randint(-5, 15)

            # Do stuff to actual mountain tiles?
            for mt in mountain["tiles"]:
                pass

            # Paint the cliff
            a = self.deftex
            b = "Cliff"
            for ct in mountain["cliff"]:
                deltex(self.hmap[ct])
                self.hmap[ct].update(self.ts.gettrans(a, b,
                                                      mountain["btmap"][ct]))
        
        # Paint water and texture features
        for feat in self.terrains:
            if len(feat["tiles"]) == 0:
                print "Ignored feature:", feat
                continue
            a = self.ts.getname(self.hmap[choose_one(feat["tiles"])]["tex"])
            b = feat["type"]
            if not b:
                b = self.ts.get_possible_transition(a)

            for ft in valid_tiles(feat["tiles"], self.mapw, self.maph):
                self.hmap[ft].update(self.ts.getrandom(b))
            for bt in valid_tiles(feat["border"], self.mapw, self.maph):
                deltex(self.hmap[bt])
                self.hmap[bt].update(self.ts.gettrans(a, b, feat["btmap"][bt]))

            print "Rendering feature '%s' on '%s'." % (b, a)

            # Apply random heights unless its a water tile
            if feat["type"] == "Water":
                for t in valid_tiles(need_water_raising(feat["tiles"]),
                                     self.mapw, self.maph):
                    self.hmap[t]["height"] = 0
            else:
                m = random.randint(5, 10) * (random.randint(0, 1)*2 - 1)
                for t in valid_tiles(need_raising(feat["tiles"]),
                                     self.mapw, self.maph):
                    self.hmap[t]["height"] += m + random.randint(-2, 2)
                    
    def save(self, name):
        ft = dict((k, self.hmap[k])
                  for k in self.valid_tiles)
        for t in ft.itervalues():
            t["height"] = clamp(t["height"], 0, 255)

        tmp = {
            "name": name,
            "tileset": self.ts.name.lower(),
            "terrain": ft,
            "width": self.mapw,
            "height": self.maph,
            "droids": self.droids,
            "features": self.features,
            "structures": self.buildings
            }

        return tmp

class TrunkTerrainPainter(TerrainPainter):
    def __init__(self, w, h, ts):
        TerrainPainter.__init__(self, w, h, ts)

    def add_water(self, tiles):
        wt = tiles | border(tiles)
        wt &= self.reachable_tiles
        self.add_texfeature(wt, "Water")
        self.paintable_tiles -= wt
        self.reachable_tiles -= wt
        self.water_tiles |= wt

    def add_texfeature(self, tiles, type=None):
        tmp = {
            "tiles": frozenset(tiles),
            "type": type,
            "border": border(tiles)
            }

        self.terrains.append(tmp)
        self.paintable_tiles -= tmp["border"]
        self.nonfinal_tiles -= tmp["border"]
        
        print "Added terrain patch with %u tiles." % len(tiles)
        
    def paint_road(self, roads, name="Road", urban=True):
        road_tiles = set()

        for road in roads:
            road_tiles |= road

        self.road_tiles |= road_tiles

    def render(self):
        # Paint water and texture features
        for feat in self.terrains:
            if len(feat["tiles"]) == 0:
                print "Ignored feature:", feat
                continue

            a = self.ts.getname(self.hmap[choose_one(feat["tiles"])]["tex"])
            b = feat["type"]
            if not b:
                b = self.ts.get_possible_transition(a)

            for ft in valid_tiles(feat["tiles"], self.mapw, self.maph):
                self.hmap[ft].update(self.ts.trunk_get_tile(b))

            # Apply random heights unless its a water tile
            if feat["type"] == "Water":
                for t in valid_tiles(need_raising(feat["tiles"]),
                                     self.mapw, self.maph):
                    self.hmap[t]["height"] = 0
            else:
                m = random.randint(5, 10) * (random.randint(0, 1)*2 - 1)
                for t in valid_tiles(need_raising(feat["tiles"]),
                                     self.mapw, self.maph):
                    self.hmap[t]["height"] += m + random.randint(-2, 2)

        if self.ts.name == "Rockies":
            for t in self.hmap:
                if t in self.water_tiles:
                    continue
                if t in self.cliff_tiles:
                    continue

                h = self.hmap[t]["height"] + random.randint(-20, 20)
                if h <= 50:
                    tex = self.ts.trunk_get_tile("Grass")
                elif h <= 100:
                    tex = self.ts.trunk_get_tile("Brown Dirt")
                elif h <= 150:
                    tex = self.ts.trunk_get_tile("Grey Dirt")
                elif h <= 200:
                    tex = self.ts.trunk_get_tile("Dirt Snow")
                else:
                    tex = self.ts.trunk_get_tile("Snow")
                self.hmap[t].update(tex)
