#------------------------------------------------------------------------------
#    Copyright (C) 2009 Florian Schanda <florian@schanda.org.uk>
#
#    This file is part of the Warzone Map Tools.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program. If not, see <http://www.gnu.org/licenses/>.
#------------------------------------------------------------------------------

from struct import pack, unpack, calcsize
import os
import zipfile
from cStringIO import StringIO

def get(fmt, f):
    return unpack(fmt, f.read(calcsize(fmt)))

def put(fmt, f, *args):
    f.write(pack(fmt, *args))

#------------------------------------------------------------------------------

const = {
    "MAX_LEVEL_SIZE": 20,

    "GTYPE_SCENARIO_START": 0,
    "GTYPE_MISSION": 2,

    # Texture attributes
    "TILE_XFLIP": 0x8000,
    "TILE_YFLIP": 0x4000,
    "TILE_ROTMASK": 0x3000,
    "TILE_ROTSHIFT": 12,
    "TILE_TRIFLIP": 0x0800,

    # Coordinates
    "TILE_SHIFT" : 7,
    "TILE_MASK" : 0x7F,
    "TILE_UNITS" : (1 << 7), #1 << TILE_SHIFT
    
    # Structure states
    "SS_BEING_BUILT" : 0,
    "SS_BUILT" : 1,
    "SS_BEING_DEMOLISHED" : 2,


    "MAX_PLAYERS" : 8,
    "MAX_SAVE_NAME_SIZE_V19": 40,
}

#------------------------------------------------------------------------------

"""
*** For the .gam file ***
Header:
  4b str  - "game"
  4b uint - version

V7 Header:
  4b uint - gameTime
  4b uint - GameType
  4b  int - ScrollMinX
  4b  int - ScrollMinY
  4b uint - ScrollMaxX
  4b uint - ScrollMaxY
  str     - levelName (len=MAX_LEVEL_SIZE)

"""

fmt_gam_header = "<4sI"
fmt_gam_v7_header = "<IIiiII%ss" % const["MAX_LEVEL_SIZE"]

def read_gam_header(fd):
    rv = {}
    rv["type"], rv["version"] = get(fmt_gam_header, fd)
    assert(rv["version"] in (7, 8))
    tmp = get(fmt_gam_v7_header, fd)
    rv["gameTime"], rv["gameType"] = tmp[0:2]
    rv["scrollMin"] = tmp[2:4]
    rv["scrollMax"] = tmp[4:6]
    rv["name"] = tmp[6]
    return rv

def write_gam_header(fd, width, height):
    put(fmt_gam_header, fd, "game", 8)
    put(fmt_gam_v7_header, fd,
        0, const["GTYPE_SCENARIO_START"],
        0, 0,
        width, height,
        "\0" * const["MAX_LEVEL_SIZE"])

#------------------------------------------------------------------------------

"""
*** For the ttypes.ttp file ***
Header:
  4b str  - "ttyp"
  4b uint - version
  4b uint - quantity

For quantity:
  2b uint - type
"""

TYPE_OF_TERRAIN = [
    "TER_SAND",
    "TER_SANDYBRUSH",
    "TER_BAKEDEARTH",
    "TER_GREENMUD",
    "TER_REDBRUSH",
    "TER_PINKROCK",
    "TER_ROAD",
    "TER_WATER",
    "TER_CLIFFFACE",
    "TER_RUBBLE",
    "TER_SHEETICE",
    "TER_SLUSH",
    
    "TER_MAX"
    ]

#------------------------------------------------------------------------------

"""
*** For the game.map file ***
Header:
  4b str  - "map "
  4b uint - version
  4b uint - width
  4b uint - height

For each tile:
  2b uint - texture
  1b uint - height

Gateway header:
  4b uint - version (should be == 1)
  4b uint - numGateways

For numGateways:
  1b uint - x0
  1b uint - y0
  1b uint - x1
  1b uint - y1

Zonemap header:
  2b uint - version (1, 2 or 3)

Zonemap header v1:
  2b uint - numZones
Zonemap header v2:
  2b uint - numZones
  2b uint - numEquivZones
  2b uint - padding?

For numZones:
  2b uint - ZoneSize


"""

fmt_header = "<4sIII"

def read_header(fd):
    rv = {}
    rv["type"], rv["version"], rv["width"], rv["height"] = get(fmt_header, fd)
    return rv

def write_header(fd, width, height):
    put(fmt_header, fd, "map ", 10, width, height)

fmt_savemap_v2 = "<HB"

def read_map(fd, width, height):
    hmap = {}
    for y in xrange(height):
        for x in xrange(width):
            tmp = {}
            foo, tmp["height"] = get(fmt_savemap_v2, fd)
            tmp["tex"] = foo & 0x01FF
            other = foo & ~0x01FF
            tmp["xflip"] = bool(other & const["TILE_XFLIP"])
            tmp["yflip"] = bool(other & const["TILE_YFLIP"])
            tmp["rot"] = ((other & const["TILE_ROTMASK"])
                          >> const["TILE_ROTSHIFT"])
            hmap[(x, y)] = tmp
    return hmap

def write_map(fd, width, height, hmap):
    for y in xrange(height):
        for x in xrange(width):
            tmp = hmap[(x, y)]
            foo = tmp["tex"]
            if tmp.get("xflip", False):
                foo |= const["TILE_XFLIP"]
            if tmp.get("yflip", False):
                foo |= const["TILE_YFLIP"]
            foo |= tmp.get("rot", 0) << const["TILE_ROTSHIFT"]
            put(fmt_savemap_v2, fd, foo, tmp["height"])


fmt_gateway_header = "<II"
fmt_gateway = "<BBBB"

def read_gateways(fd):
    GW_VERSION, GW_NUM = get(fmt_gateway_header, fd)
    assert(GW_VERSION == 1)

    gw = []
    for i in xrange(GW_NUM):
        gw.append(get(fmt_gateway, fd))

    return gw

def write_gateways(fd, gw):
    put(fmt_gateway_header, fd, 1, len(gw))
    for x in gw:
        assert(len(x) == 4)
        put(fmt_gateway, fd, *x)

fmt_zonemap_version = "<H"
fmt_zonemap_header = {
    1: "<H",
    2: "<HHH"
}

def read_zones(fd):
    ZM_VERSION, = get(fmt_zonemap_version, fd)

    if ZM_VERSION == 1:
        ZM_NUM, = get(fmt_zonemap_header[1], fd)
    elif ZM_VERSION == 2:
        ZM_NUM, ZM_NUMEQUIV, dummy = get(fmt_zonemap_header[2], fd)
    else:
        print "Unknown zonemap version = %u." % ZM_VERSION

    zones = []
    for i in xrange(ZM_NUM):
        zone_size, = get("<H", fd)
        z = []
        for j in xrange(zone_size):
            z += get("<B", fd)
        zones.append(z)

    #if ZM_VERSION == 2:
    #    print ZM_NUMEQUIV
    #    for i in xrange(ZM_NUMEQUIV):
    #        print i, get("<I", fd)

    return zones

def write_zones(fd, zones):
    put(fmt_zonemap_version, fd, 1)
    put(fmt_zonemap_header[1], fd, len(zones))
    for z in zones:
        put("<H", fd, len(z))
        for i in z:
            put("<B", fd, i)

#------------------------------------------------------------------------------

""" feat.bjo - we cater to version 8 exactly
Header:
  4b str  - "feat"
  4b uint - version
  4b uint - quantity

(OBJECT_SAVE_V19) / (SAVE_FEATURE_V2)
  string  - name (len = MAX_SAVE_NAME_SIZE_V19)
  4b uint - id
  4b uint - x, y, z (z can be ignored for v14 features)
  4b uint - direction
  4b uint - player (this can be ignored for v14 features)
  4b bool - inFire
  4b uint - burnStart (this can be ignored for v14 features)
  4b uint - burnDamage
(FEATURE_SAVE_V14) / (SAVE_FEATURE_V14)
  1b uint - visible[max_players]
"""

fmt_feature_header = "<4sII"
fmt_feature_v2 = "<%ssIIIIIIIII" % const["MAX_SAVE_NAME_SIZE_V19"]

def read_features(fd):
    F_TYPE, F_VERSION, F_NUM = get(fmt_feature_header, fd)
    assert(F_TYPE == "feat")
    assert(F_VERSION <= 19)
    assert(F_VERSION == 8)

    rv = []
    for i in xrange(F_NUM):
        f = {}
        tmp = get(fmt_feature_v2, fd)
        f["name"] = tmp[0].split("\0", 2)[0]
        f["pos"] = [x >> const["TILE_SHIFT"] for x in tmp[2:4]] # Ignoring Z
        f["tpos"] = [x & const["TILE_MASK"] for x in tmp[2:4]] # Ignoring Z
        f["dir"] = tmp[5]
        f["inFire"] = bool(tmp[7])
        f["burnDamage"] = tmp[9]
        rv.append(f)
    return rv

def write_features(fd, ids, feat):
    put(fmt_feature_header, fd, "feat", 8, len(feat))
    for f in feat:
        put(fmt_feature_v2, fd,
            f["name"],
            ids.next(),
            (f["pos"][0] << const["TILE_SHIFT"]) + f.get("tpos", (0, 0))[0],
            (f["pos"][1] << const["TILE_SHIFT"]) + f.get("tpos", (0, 0))[1],
            0,
            f.get("dir", 0),
            0,
            f.get("inFire", 0),
            0, f.get("burnDamage", 0))

#------------------------------------------------------------------------------

""" For dinit.bjo - cater for v8?
Header:
  4b str  - "dint"
  4b uint - version
  4b uint - quantity

For quantity:
(OBJECT_SAVE_V19) / (SAVE_DROIDINIT)
  string  - name (len = MAX_SAVE_NAME_SIZE_V19)
  4b uint - id
  4b uint - x, y, z (z is ignored)
  4b uint - direction
  4b uint - player
  4b bool - inFire (ignored)
  4b uint - burnStart (ignored)
  4b uint - burnDamage (ignored)
"""

fmt_droid_header = "<4sII"
fmt_save_droidinit = "<%usIIIIIIIII" % const["MAX_SAVE_NAME_SIZE_V19"]

def read_droids(fd):
    D_TYPE, D_VERSION, D_NUM = get(fmt_droid_header, fd)
    assert(D_TYPE == "dint")
    assert(D_VERSION >= 7)

    rv = []
    for i in xrange(D_NUM):
        f = {}
        tmp = get(fmt_save_droidinit, fd)
        f["name"] = tmp[0].split("\0", 2)[0]
        f["pos"] = [x >> const["TILE_SHIFT"] for x in tmp[2:4]] # Ignoring Z
        f["dir"] = tmp[5]
        f["player"] = tmp[6]
        rv.append(f)
    return rv

def write_droids(fd, ids, droids):
    put(fmt_droid_header, fd, "dint", 7, len(droids))
    for d in droids:
        put(fmt_save_droidinit, fd,
            d["name"],
            ids.next(),
            d["pos"][0] << const["TILE_SHIFT"],
            d["pos"][1] << const["TILE_SHIFT"],
            0,
            d.get("dir", 0), d["player"],
            0, 0, 0)

#------------------------------------------------------------------------------

""" For struct.bjo
Header:
  4b str  - "stru"
  4b uint - version
  4b uint - quantity

For quantity:
(SAVE_STRUCTURE_V2) (STRUCTURE_SAVE_V2)
  (OBJECT_SAVE_V19)
 0    string  - name (len = MAX_SAVE_NAME_SIZE_V19)
 1    4b uint - id
 2    4b uint - x, y, z
 5    4b uint - direction
 6    4b uint - player
 7    4b bool - inFire
 8    4b uint - burnStart
 9    4b uint - burnDamage
------------------------------
10  1b uint - status (but its really 4b long!!!)
11  4b  int - currentBuildPts
12  4b uint - body
13  4b uint - armour
14  4b uint - resistance
15  4b uint - dummy1
16  4b uint - subjectInc;  /*research inc or factory prod id*/\
17  4b uint - timeStarted
18  4b uint - output
19  4b uint - capacity
20  4b uint - quantity
"""

fmt_object_save_v19 = "<%usIIIIIIIII" % const["MAX_SAVE_NAME_SIZE_V19"]
fmt_save_structure_v2 = fmt_object_save_v19 + "BxxxiIIIIIIIII"
fmt_struct_header = "<4sII"

def read_structs(fd):
    S_TYPE, S_VERSION, S_NUM = get(fmt_struct_header, fd)
    print "Struct:", S_TYPE, S_VERSION, S_NUM
    assert(S_TYPE == "stru")
    assert(S_VERSION >= 7)
    assert(S_VERSION == 8)

    rv = []
    for i in xrange(S_NUM):
        f = {}
        tmp = get(fmt_save_structure_v2, fd)
        f["name"] = tmp[0].split("\0", 2)[0]
        f["pos"] = [x >> const["TILE_SHIFT"] for x in tmp[2:4]] # Ignoring Z
        f["dir"] = tmp[5]
        f["player"] = tmp[6]
        rv.append(f)
    return rv

def write_structs(fd, ids, structs):
    put(fmt_struct_header, fd, "stru", 8, len(structs))
    for s in structs:
        put(fmt_save_structure_v2, fd,
            s["name"],
            ids.next(),
            s["pos"][0] << const["TILE_SHIFT"],
            s["pos"][1] << const["TILE_SHIFT"],
            0,
            s.get("dir", 0), s["player"],
            s.get("inFire", False),
            s.get("burnStart", 0),
            s.get("burnDamage", 0),
            s.get("status", const["SS_BUILT"]),
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0)


#------------------------------------------------------------------------------

def install_map(map):
    # Get implicit information about map
    hmap = map["terrain"]
    name = map["name"]
    ts   = map["tileset"]
    droids = map["droids"]
    features = map.get("features", [])
    structures = map.get("structures", [])

    players = len(frozenset(x["player"]
                            for x in droids
                            if x.get("player", None) != None))
    max_x, max_y = [max(x) + 1 for x in zip(*hmap.keys())]

    if players not in (2, 4, 8):
        raise ValueError, "Number of players must be 2, 4 or 8. (is %u!)" % \
            players

    if ts == "arizona":
        ts_id = 1
    elif ts == "urban":
        ts_id = 2
    elif ts == "rockies":
        ts_id = 3
    else:
        raise ValueError, "Invalid tileset: '%s'." % ts

    # Set up file names
    bdir = os.path.join(os.path.expanduser("~"), ".warzone2100-2.2")

    bname = ("%uc-%s" % (players, name)).lower()
    lev_file = os.path.join("%s.addon.lev" % bname)
    gam_file = os.path.join("multiplay", "maps", "%s.gam" % bname)
    map_dir  = os.path.join("multiplay", "maps", bname)

    # Make sure directories exist
    if not os.path.isdir(bdir):
        raise IOError, "Could not find '%s'." % bdir

    for d in [os.path.join(bdir, "maps")]:
        if not os.path.isdir(d):
            if not os.path.exists(d):
                os.mkdir(d)
            else:
                raise IOError, "Could not create directory '%s'." % d

    wz_file = os.path.join(bdir, "maps", "%s.wz" % bname)
    zf = zipfile.ZipFile(wz_file, "w", zipfile.ZIP_DEFLATED)
    
    # Write map
    def idgen():
        id = 0
        while True:
            yield id
            id += 1

    ids = idgen()

    print "Writing %ux%u map '%s'" % (max_x, max_y, wz_file)

    out = StringIO()
    write_gam_header(out, max_x, max_y)
    zf.writestr(gam_file, out.getvalue())
    out.close()

    tech_game_type_map = {
        (1, 0): 12,
        (1, 1): 14,
        (2, 0): 15,
        (3, 0): 16,
        (2, 1): 18,
        (3, 1): 19
        }
    out = StringIO()
    for sk in xrange(2):
        for tl in xrange(1, 3+1):
            # Create correct level name
            lname = "Sk-" if sk else ""
            lname += name
            lname += "" if tl == 1 else "-T%u" % tl
            out.write("level %s\n" % lname)

            # Player number
            out.write("players %u\n" % players)

            # Some magic type flag
            out.write("type %u\n" % tech_game_type_map[(tl, sk)])

            # Dataset
            if tl == 1:
                out.write("dataset MULTI_CAM_%u\n" % ts_id)
            else:
                out.write("dataset MULTI_T%u_C%u\n" % (tl, ts_id))

            # "Savegame"
            out.write("game \"multiplay/maps/%s.gam\"\n" % bname)

            # Resources
            resname = "skirmish%u" % players if sk else "campaign"
            if sk and tl > 1:
                resname = ("t%u-" % tl) + resname
            out.write("data \"wrf/multi/%s.wrf\"\n" % resname)

            # "Fog"
            out.write("data \"wrf/multi/fog%u.wrf\"\n" % ts_id)
            out.write("\n")
    zf.writestr(lev_file, out.getvalue())
    out.close()

    out = StringIO()
    write_header(out, max_x, max_y)
    write_map(out, max_x, max_y, hmap)
    write_gateways(out, [])
    write_zones(out, [])
    zf.writestr(os.path.join(map_dir, "game.map"), out.getvalue())
    out.close()

    # Below for the experimental ascii map format
    # out = open("map.ascii", "w")
    # out.write("WIDTH %u\n" % max_x)
    # out.write("HEIGHT %u\n" % max_y)
    # for y in xrange(max_y):
    #     for x in xrange(max_x):
    #         out.write("X %u Y %u H %u TEX %u ROT %u XFLIP %u YFLIP %u\n" %
    #                   (x, y, hmap[(x, y)]["height"],
    #                    hmap[(x, y)]["tex"],
    #                    hmap[(x, y)].get("rot", 0),
    #                    hmap[(x, y)].get("xflip", 0),
    #                    hmap[(x, y)].get("yflip", 0)))
    # out.close()

    out = StringIO()
    write_droids(out, ids, droids)
    zf.writestr(os.path.join(map_dir, "dinit.bjo"), out.getvalue())
    out.close()

    out = StringIO()
    write_features(out, ids, features)
    zf.writestr(os.path.join(map_dir, "feat.bjo"), out.getvalue())
    out.close()

    out = StringIO()
    write_structs(out, ids, structures)
    zf.writestr(os.path.join(map_dir, "struct.bjo"), out.getvalue())
    out.close()

    zf.write("tilesets/%s.ttp" % ts, os.path.join(map_dir, "ttypes.ttp"))

    zf.close()
