// 
// Copyright (c) 2006-2007, Benjamin Kaufmann
// 
// This file is part of Clasp. See http://www.cs.uni-potsdam.de/clasp/ 
// 
// Clasp is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// Clasp is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Clasp; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//
#include <clasp/include/lparse_reader.h>
#include <clasp/include/satelite.h>
#include <clasp/include/unfounded_check.h>
#include <clasp/include/cb_enumerator.h>
#include <clasp/include/model_enumerators.h>
#include <clasp/include/smodels_constraints.h>
#include <clasp/app/timer.h>
#include "options.h"
#include <iostream>
#include <iomanip>
#include <string>
#include <fstream>
#include <numeric>
#include <signal.h>

#if defined (_MSC_VER) && _MSC_VER >= 1200
//#define CHECK_HEAP 1;
#include <crtdbg.h>
#endif

using namespace std;
using namespace Clasp;

class StdOutPrinter : public ModelPrinter, public CBConsequences::CBPrinter {
public:
	explicit StdOutPrinter(bool quiet) : quiet_(quiet) {}
	void printModel(const Solver& s, const ModelEnumerator& e) {
		if (quiet_) return;
		AtomIndex* index = e.getIndex();
		if (index) {
			std::cout << "Answer: " << s.stats.search.models << "\n";
			for (AtomIndex::size_type i = 0; i != index->size(); ++i) {
				if (s.value((*index)[i].lit.var()) == trueValue((*index)[i].lit) && !(*index)[i].name.empty()) {
					std::cout << (*index)[i].name << " ";
				}
			}
			std::cout << std::endl;
		}
		else {
			const uint32 numVarPerLine = 10;
			std::cout << "c Model: " << s.stats.search.models << "\n";
			std::cout << "v ";
			for (Var v = 1, cnt=0; v <= s.numVars(); ++v) {
				if (s.value(v) == value_false) std::cout << "-";
				std::cout << v;
				if (++cnt == numVarPerLine && v+1 <= s.numVars()) { cnt = 0; std::cout << "\nv"; }
				std::cout << " ";
			}
			std::cout << "0 \n" << std::flush;
		}
		printMini(e.getMini(), index?"Optimization":"c Optimization");
	}
	void printReport(const Solver& s, const ModelEnumerator& e) {
		printModels(s, e.getIndex() ? "Models" : "c Models", e.getMini(), 0);
		if (quiet_) {
			printMini(e.getMini(), e.getIndex()?"Optimization":"c Optimization");
		}
	}
	void printConsequences(const Solver&, const CBConsequences& cb) {
		if (!quiet_) {
			std::cout << (cb.cbType() == CBConsequences::cautious_consequences ? "Cautious" : "Brave") << " consequences: \n";
			const AtomIndex& index = cb.index();
			for (AtomIndex::size_type i = 0; i != index.size(); ++i) {
				if (index[i].lit.watched()) {
					std::cout << index[i].name << " ";
				}
			}
			std::cout << std::endl;
			printMini(cb.mini(), "Optimization");
		}
	}
	void printReport(const Solver& s, const CBConsequences& cb) {
		if (quiet_ && s.stats.search.models > 0) {
			quiet_ = false;
			printConsequences(s, cb);
		}
		printModels(s, "Models", cb.mini(), &cb);
		if (quiet_) printMini(cb.mini(), "Optimization");
	}
	void printMini(MinimizeConstraint* c, const char* prefix) {
		if (!c || c->numRules() == 0 || c->models() == 0) return;
		std::cout << std::left << std::setw(12) << prefix << ":";
		for (uint32 i = 0; i < c->numRules(); ++i) {
			std::cout << " " << c->getOptimum(i);
		}
		std::cout << std::endl;
	}
private:
	void printModels(const Solver& s, const char* str, MinimizeConstraint* min, const CBConsequences* cons) {
		uint64 enumerated = s.stats.search.models;
		uint64 models;
		if      (cons) models = std::min(enumerated, uint64(1));
		else if (min)  models = min->models();
		else           models = enumerated;
		std::cout << "\n";
		std::cout << std::left << std::setw(12) << str << ": ";
		bool ok = true;
		if (s.decisionLevel() != s.rootLevel()) {
			char buf[64];
			int wr    = sprintf(buf, "%llu", models);
			buf[wr]   = '+';
			buf[wr+1] = 0;
			std::cout << std::setw(6) << buf;	
			ok = false;
		}
		else {
			std::cout << std::setw(6) << models;
		}
		cout << "\n";
		if (enumerated != models) {
			std::cout << std::left << std::setw(12) << "  Enumerated" << ": " << enumerated << "\n";
		}
		if (min && models) {
			std::cout << std::left << std::setw(12) << "  Optimum" << ": " << (ok?"yes":"unknown") << "\n";
		}
		if (cons && models) {
			std::cout << std::left << std::setw(12) << (cons->cbType() == CBConsequences::cautious_consequences ? "  Cautious" : "  Brave")
			          << ":" << (ok?"yes":"unknown") << "\n";
		}
		std::cout << std::flush;
	}

	bool quiet_;
};

struct ClaspApp {
	ClaspApp() : mini_(0) {}
	Options       options;
	Solver        solver;
	Timer         t_pre;
	Timer         t_solve;
	Timer         t_all;
	uint32        problemSize;
	enum Mode {
		ASP_MODE,
		SAT_MODE
	}             mode;
	int   run(int argc, char** argv);
	void  printKillStats();
private:
	enum State { start_read, start_pre, start_solve, end_read, end_pre, end_solve };
	void  setState(State s);
	void  printSatStats() const;
	void  printLpStats()  const;
	void  printAspStats() const;
	std::string shortName(const std::string& str) {
		if (str.size() < 40) return str;
		std::string r = "...";
		r.append(str.end()-38, str.end());
		return r;
	}
	static void printSatEliteProgress(uint32 i, SatElite::SatElite::Options::Action a) {
		static const char pro[] = {'/', '-', '\\', '|' };
		if      (a==SatElite::SatElite::Options::pre_start)   cout << '|';
		else if (a==SatElite::SatElite::Options::iter_start)  cout << '\b' << pro[i%4];
		else if (a==SatElite::SatElite::Options::pre_done)    cout << '\b' << "Done";
		else if (a==SatElite::SatElite::Options::pre_stopped) cout << '\b' << "Stop";
	}
	std::istream& getStream() {
		if (!options.file.empty()) {
			if (inFile_.is_open()) return inFile_;
			inFile_.open(options.file.c_str());
			if (!inFile_) {
				throw std::runtime_error("Can not read from '"+options.file+"'"); 
			}
			return inFile_;
		}
		return std::cin;
	}
	bool solve();
	bool parseLparse();
	bool readSat();
	std::auto_ptr<PreproStats>  preStats_;
	std::auto_ptr<AtomIndex>    atomIndex_;
	Enumerator*                 enum_;
	MinimizeConstraint*         mini_;
	std::ifstream inFile_;
} clasp_g;



ostream& operator<<(ostream& os, const Timer& t) {
	streamsize old = os.precision(3);
	os << fixed << t.total();
	os.precision(old);
	return os;
}

// return values
const int S_UNKNOWN       = 0;
const int S_SATISFIABLE   = 10;
const int S_UNSATISFIABLE = 20;
const int S_ERROR         = EXIT_FAILURE;
const int S_MEMORY        = 107;

static void sigHandler(int) {
	// ignore further signals
	signal(SIGINT, SIG_IGN);    // Ctrl + C
	signal(SIGTERM, SIG_IGN);   // kill (but not kill -9)
	clasp_g.t_all.stop();
	clasp_g.printKillStats();
	_exit(clasp_g.solver.stats.search.models > 0 ? S_SATISFIABLE : S_UNKNOWN);
}

void ClaspApp::setState(State s) {
	bool q = options.quiet;
	const int width = 13;
	switch (s) {
		case start_read:
			t_pre.reset(); t_pre.start();
			if (!q) cout << (mode == ASP_MODE ?"":"c ") << left << setw(width) << "Reading" << ": ";
			break;
		case end_read:
			t_pre.stop();
			if (!q) cout << "Done(" << t_pre << "s)\n";
			break;
		case start_pre:
			t_pre.reset(); t_pre.start();
			if (!q) cout << (mode == ASP_MODE ?"":"c ") << left << setw(width) << "Preprocessing" << ": ";
			break;
		case end_pre:
			t_pre.stop();
			if (!q) cout << (options.satPreParams[0]==0?"Done(":"(") << t_pre << "s)\n";
			break;
		case start_solve:
			t_solve.start();
			cout << (mode == ASP_MODE ?"":"c ") << "Solving...\n";
			break;
		case end_solve:
			t_solve.stop();
			break;
	default:;
	}
}

void ClaspApp::printSatStats() const {
	const SearchStats& st  = solver.stats.search;
	const ProblemStats& ps = solver.stats.problem;
	cout << "c Time         : " << t_all << "s (Solve: " << t_solve <<"s)\n";
	if (!options.stats) return;
	cout << "c Choices      : " << st.choices << "\n";
	cout << "c Conflicts    : " << st.conflicts <<"\n";
	cout << "c Restarts     : " << st.restarts << "\n";
	cout << "c Variables    : " << solver.numVars() << " (Eliminated: " << solver.numEliminatedVars() << ")\n";
	cout << "c Clauses      : " << ps.constraints[0] << "\n";
	cout << "c   Binary     : " << ps.constraints[1] << "\n";
	cout << "c   Ternary    : " << ps.constraints[2] << "\n";
	cout << "c Lemmas       : " << st.learnt[0] << "\n";
	cout << "c   Binary     : " << st.learnt[1] << "\n";
	cout << "c   Ternary    : " << st.learnt[2] << "\n";
	cout << "c   Deleted    : " << st.learnt[3] << "\n";
}

void ClaspApp::printKillStats() {
	const char* pre = mode == ClaspApp::SAT_MODE ? "c " : "";
	printf("\n%s*** INTERRUPTED! ***\n", pre);
	printf("%s%-12s: %u\n"  , pre, "Models"   , (uint32)clasp_g.solver.stats.search.models);  	
	if (mini_ && mini_->models() > 0) {
		printf("%s%-12s:", pre, "Optimization");
		for (uint32 i = 0; i < mini_->numRules(); ++i) {
			printf(" %d", mini_->getOptimum(i));
		}
		printf("\n");
	}
	printf("%s%-12s: %.3f\n", pre, "Time"     , clasp_g.t_all.total());  
	printf("%s%-12s: %u\n"  , pre, "Choices"  , (uint32)clasp_g.solver.stats.search.choices); 
	printf("%s%-12s: %u\n"  , pre, "Conflicts", (uint32)clasp_g.solver.stats.search.conflicts); 
	printf("%s%-12s: %u\n"  , pre, "Restarts" , (uint32)clasp_g.solver.stats.search.restarts);
	printf("%s%s\n"         , pre, clasp_g.solver.stats.search.models > 0 ? "SATISFIABLE" : "UNKNOWN");
}

static double percent(uint32 y, uint32 from) {
	if (from == 0) return 0;
	return (static_cast<double>(y)/from)*100.0;
}
static double percent(const uint32* arr, uint32 idx) {
	return percent(arr[idx], arr[0]);
}
static double compAverage(uint64 x, uint64 y) {
	if (!x || !y) return 0.0;
	return static_cast<double>(x) / static_cast<double>(y);
}

void ClaspApp::printLpStats() const {
	const Options& o = options;
	PreproStats& ps  = *preStats_;
	if (ps.trStats) {
	  ps.rules[BASICRULE] -= ps.trStats->rules[0];
	  for (uint32 rt = BASICRULE+1; rt <= OPTIMIZERULE; ++rt) {
			ps.rules[0] -= ps.trStats->rules[rt];
		}
	}
	cout << left << setw(12) << "Atoms" << ": " << setw(6) << ps.atoms;
	if (ps.trStats) {
		cout << " (Original: " << ps.atoms-ps.trStats->auxAtoms << " Auxiliary: " << ps.trStats->auxAtoms << ")";
	}
	cout << "\n";
	cout << left << setw(12) << "Rules" << ": " << setw(6) << ps.rules[0] << " (";
	for (int i = 1; i <= OPTIMIZERULE; ++i) {
		if (ps.rules[i] != 0) {
			cout << i << ": " << ps.rules[i];
			if (ps.trStats && ps.trStats->rules[0] != 0) {
				if (i == BASICRULE) {
					cout << "/" << ps.rules[BASICRULE]+ps.trStats->rules[0];
				}
				else {
					cout << "/" << ps.rules[i]-ps.trStats->rules[i];
				}
			}
			cout << " ";
		}
	}
	cout << "\b)" << "\n";
	cout << left << setw(12) << "Bodies" << ": " << ps.bodies << "\n";
	if (ps.numEqs() != 0) {
		cout << left << setw(12) << "Equivalences" << ": " << setw(6) << ps.numEqs() << " (Atom=Atom: " << ps.numEqs(Var_t::atom_var)
				 << " Body=Body: "   << ps.numEqs(Var_t::body_var)
				 << " Other: " << (ps.numEqs() - ps.numEqs(Var_t::body_var) - ps.numEqs(Var_t::atom_var))
				 << ")" << "\n";
	} 
	cout << left << setw(12) << "Tight" << ": ";
	if (!o.suppModels) {
		if (ps.sccs == 0) {
			cout << "Yes";
		}
		else {
			cout << setw(6) << "No"  << " (SCCs: " << ps.sccs << " Nodes: " 
					 << ps.ufsNodes << ")";
		}
	}
	else {
		cout << "N/A";
	}
	cout << "\n"; 
}

void ClaspApp::printAspStats() const {
	cout.precision(1);
	cout.setf(ios_base::fixed, ios_base::floatfield);
	const SearchStats& st      = solver.stats.search;
	const JumpStats& js        = solver.stats.jump;
	const ProblemStats& ps     = solver.stats.problem;
	cout  << left << setw(12) << "Time"  << ": " << setw(6) << t_all
				<< " (" << "Solving: " << t_solve << ")" << "\n";
	if (!options.stats) return;
	cout << left << setw(12) << "Choices"       << ": " << st.choices   << "\n";
	cout << left << setw(12) << "Conflicts"     << ": " << st.conflicts << "\n";
	cout << left << setw(12) << "Restarts"      << ": " << st.restarts  << "\n";
	cout << "\n";
	printLpStats();
	cout << "\n";
	uint32 other = ps.constraints[0] - ps.constraints[1] - ps.constraints[2];
	cout << left << setw(12) << "Variables" << ": " << setw(6) << solver.numVars() 
				<< " (Eliminated: " << right << setw(4) << solver.numEliminatedVars() << ")" << "\n";
	cout  << left << setw(12) << "Constraints" << ": " << setw(6) << ps.constraints[0]
				<< " (Binary: " << right << setw(4) << percent(ps.constraints, 1) <<  "% "
				<< "Ternary: "  << right << setw(4) << percent(ps.constraints, 2) <<  "% "
				<< "Other: "    << right << setw(4) << percent(other, ps.constraints[0]) << "%)"
				<< "\n";
	other = st.learnt[0] - st.learnt[1] - st.learnt[2];
	cout  << left << setw(12) << "Lemmas" << ": " << setw(6) << st.learnt[0]
				<< " (Binary: " << right << setw(4) << percent(st.learnt, 1) << "% "
				<< "Ternary: "  << right << setw(4) << percent(st.learnt, 2) <<  "% "
				<< "Other: "    << right << setw(4) << percent(other, st.learnt[0]) << "%)"
				<< "\n";
	cout << left << setw(12) << "  Conflicts"     << ": " << setw(6) << st.learnt[0]-st.loops 
				<< " (Average Length: " << compAverage(st.lits[0], st.learnt[0]-st.loops) << ")\n";
	cout << left << setw(12) << "  Loops" << ": " << setw(6) << st.loops 
				<< " (Average Length: " << compAverage(st.lits[1], st.loops) << ")\n";
#if MAINTAIN_JUMP_STATS == 1
	cout << "\n";
	cout << "Backtracks          : " << st.conflicts-js.jumps << "\n";
	cout << "Backjumps           : " << js.jumps << " ( Bounded: " << js.bJumps << " )\n";
	cout << "Skippable Levels    : " << js.jumpSum << "\n"; 
	cout << "Levels skipped      : " << js.jumpSum - js.boundSum << " ("
			<< 100*((js.jumpSum - js.boundSum)/std::max(1.0, (double)js.jumpSum)) << "%)" << "\n";
	cout << "Max Jump Length     : " << js.maxJump << " ( Executed: " << js.maxJumpEx << " )\n";
	cout << "Max Bound Length    : " << js.maxBound << "\n";
	cout << "Average Jump Length : " << js.avgJumpLen() << " ( Executed: " << js.avgJumpLenEx() << " )\n";
	cout << "Average Bound Length: " << (js.boundSum)/std::max(1.0,(double)js.bJumps) << "\n";
	cout << "Average Model Length: " << (js.modLits)/std::max(1.0, (double)st.models) << "\n";
#endif
	cout << endl;
}

int ClaspApp::run(int argc, char** argv) {
	if (!options.parse(argc, argv, std::cout, solver)) {
		throw std::runtime_error( options.getError() );
	}
	if (!options.getWarning().empty()) {
		cerr << options.getWarning() << endl;
	}
	if (options.help || options.version) {
		return EXIT_SUCCESS;
	}
	if (options.seed != -1) { Clasp::srand(options.seed); }
	mode = options.dimacs ? SAT_MODE : ASP_MODE;
	if (options.satPreParams[0] != 0) {
		// enable and configure the sat preprocessor
		SatElite::SatElite* pre = new SatElite::SatElite(solver);
		pre->options.maxIters = options.satPreParams[0];
		pre->options.maxOcc   = options.satPreParams[1];
		pre->options.maxTime  = options.satPreParams[2];
		pre->options.elimPure = options.numModels == 1;
		pre->options.verbose  = options.quiet ? 0 : printSatEliteProgress;
		solver.strategies().satPrePro.reset(pre);
	}
	cout << (mode == ASP_MODE ?"":"c ") << "clasp version " << VERSION << " \n";
	getStream();
	cout << (mode == ASP_MODE ?"":"c ") << "Reading from " << (options.file.empty()?"stdin":shortName(options.file).c_str()) << "\n";
	t_all.start();
	solve();
	t_all.stop();
	if (mode == ASP_MODE) {
		printAspStats();
	}
	else {
		cout << "s " << (solver.stats.search.models != 0 ? "SATISFIABLE" : "UNSATISFIABLE") << "\n";
		printSatStats();
	}
	return solver.stats.search.models != 0 ? S_SATISFIABLE : S_UNSATISFIABLE;
}

bool ClaspApp::solve() {
	bool res = mode == ASP_MODE ? parseLparse() : readSat();
	assert(enum_);
	enum_->setNumModels(options.numModels!=-1?options.numModels:1);
	solver.add(enum_);
	options.solveParams.setEnumerator( *enum_ );
	if (res) {
		if (options.redFract == 0.0) {
			options.solveParams.setReduceParams((uint32)-1, 1.0, (uint32)-1, options.redOnRestart);
		}
		else {
			options.solveParams.setReduceParams(static_cast<uint32>(problemSize/options.redFract)
				, options.redInc
				, static_cast<uint32>(problemSize*options.redMax)
				, options.redOnRestart
			);
		}
		setState(start_solve);
		res = Clasp::solve(solver, options.solveParams);
		setState(end_solve);
	}
	else {
		enum_->endSearch(solver);
	}
	return res;
}


bool ClaspApp::readSat() {
	std::istream& in = getStream();
	enum_ = options.recordSol 
		? (Enumerator*)new RecordEnumerator(new StdOutPrinter(options.quiet), options.modelRestart)
		: (Enumerator*)new BacktrackEnumerator(new StdOutPrinter(options.quiet), options.projectConfig);
	setState(start_read);
	bool res = parseDimacs(in, solver, options.numModels == 1);
	setState(end_read);
	if (res) {
		setState(start_pre);
		res = solver.endAddConstraints();
		setState(end_pre);
	}
	problemSize = solver.numConstraints();
	return res;
}

bool ClaspApp::parseLparse() {
	std::istream& in = getStream();
	preStats_.reset(new PreproStats());
	atomIndex_.reset(new AtomIndex());
	LparseReader reader;
	ProgramBuilder api;
	api.setExtendedRuleMode(ProgramBuilder::ExtendedRuleMode(options.transExt));
	api.startProgram(*atomIndex_, options.suppModels 
		? 0
		: new DefaultUnfoundedCheck(DefaultUnfoundedCheck::ReasonStrategy(options.loopRep))
		, (uint32)options.eqIters
	);
	setState(start_read);
	bool ret = reader.parse(in, api);
	setState(end_read);
	if (options.numModels == -1) {
		options.numModels = api.hasMinimize() == false;
	}
	if ( (api.hasMinimize() || !options.cons.empty()) && options.numModels != 0) {
		if (api.hasMinimize()) {
			cerr << "Warning: Option '--number' not 0: last model may not be optimal!" << endl;
		}
		if (!options.cons.empty()) {
			cerr << "Warning: Option '--number' not 0: last model may not cover all consequences!" << endl;
		}
	}
	if (api.hasMinimize() && !options.cons.empty()) {
		cerr << "Warning: Minimize statements: Consequences may depend on enumeration order!" << endl;
	}
	if (api.hasMinimize() && options.project) {
		cerr << "Warning: Projection: Optimization values may depend on enumeration order!" << endl;
	}
	mini_ = 0;
	if (ret) {
		setState(start_pre);
		ret = api.endProgram(solver, false);
		api.stats.moveTo(*preStats_);
		if (ret && api.hasMinimize()) {
			mini_ = api.createMinimize(solver);
			mini_->setMode( options.optAll ? MinimizeConstraint::compare_less_equal : MinimizeConstraint::compare_less );
			if (!options.optVals.empty()) {
				uint32 m = std::min((uint32)options.optVals.size(), mini_->numRules());
				for (uint32 x = 0; x != m; ++x) {
					mini_->setOptimum(x, options.optVals[x]);
				}
			}
		}
	}
	if (!options.cons.empty()) {
		enum_ = new CBConsequences(solver, atomIndex_.get(), 
			options.cons == "brave" ? CBConsequences::brave_consequences : CBConsequences::cautious_consequences, mini_,
			new StdOutPrinter(options.quiet), options.modelRestart
		);
	}
	else {
		ModelEnumerator* e = !options.recordSol 
			? (ModelEnumerator*)new BacktrackEnumerator(new StdOutPrinter(options.quiet), options.projectConfig)
			: (ModelEnumerator*)new RecordEnumerator(new StdOutPrinter(options.quiet), options.modelRestart);
		e->startSearch(solver, atomIndex_.get(), options.project, mini_);
		enum_ = e;
	}
	ret = ret && solver.endAddConstraints();
	double r = solver.numVars() / double(solver.numConstraints());
	if (r < 0.1 || r > 10.0) {
		problemSize = std::max(solver.numVars(), solver.numConstraints());
	}
	else {
		problemSize = std::min(solver.numVars(), solver.numConstraints());
	}
	setState(end_pre);
	return ret;
}

int main(int argc, char** argv) {
#if defined (_MSC_VER) && defined (CHECK_HEAP) && _MSC_VER >= 1200 
	_CrtSetDbgFlag(_CrtSetDbgFlag(_CRTDBG_REPORT_FLAG) |
					_CRTDBG_LEAK_CHECK_DF | _CRTDBG_ALLOC_MEM_DF |
	_CRTDBG_CHECK_ALWAYS_DF);

	_CrtSetReportMode( _CRT_WARN, _CRTDBG_MODE_FILE );
	_CrtSetReportFile( _CRT_WARN, _CRTDBG_FILE_STDERR );
	_CrtSetReportMode( _CRT_ERROR, _CRTDBG_MODE_FILE );
	_CrtSetReportFile( _CRT_ERROR, _CRTDBG_FILE_STDERR );
	_CrtSetReportMode( _CRT_ASSERT, _CRTDBG_MODE_FILE );
	_CrtSetReportFile( _CRT_ASSERT, _CRTDBG_FILE_STDERR );
#endif
	try {
		signal(SIGINT, sigHandler);    // Ctrl + C
		signal(SIGTERM, sigHandler);   // kill (but not kill -9)
		return clasp_g.run(argc, argv);
	} 
	catch(const ReadError& e) {
		cerr << "Failed!\nInput Error at line " << e.line_ << ": " << e.what() << endl;
		return S_ERROR;
	}
	catch (const std::bad_alloc&) {
		cerr << "\nclasp ERROR: out of memory"  << endl;
		return S_MEMORY;
	}
	catch (const std::exception& e) {
		cerr << "\nclasp ERROR: " << e.what() << endl;
		return S_ERROR;
	}
	return S_UNKNOWN;
}
