#------------------------------------------------------------------------------
#    Copyright (C) 2009 Florian Schanda <florian@schanda.org.uk>
#
#    This file is part of the Warzone Map Tools.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program. If not, see <http://www.gnu.org/licenses/>.
#------------------------------------------------------------------------------

import random

cdef extern from "stdlib.h":
    void *malloc(int size)
    void free(void *ptr)
    int rand()
    void srand(unsigned int seed)

cdef extern from "sorted_list.h":
    struct SortedList:
        pass

    SortedList * new_slist()
    void delete_slist(SortedList *l)

    void slist_add(SortedList *l, int coord, int cost)
    void slist_pop(SortedList *l)
    int slist_head_coord(SortedList *l)
    int slist_head_cost(SortedList *l)
    int slist_is_empty(SortedList *l)


def route(p_start, p_end, set_valid, set_cheap):
    srand(random.getrandbits(31))
    
    cdef int sx, sy, x, y, i
    
    sx = 0
    sy = 0
    for x, y in set_valid:
        if x > sx:
            sx = x
        if y > sy:
            sy = y
    sx += 1
    sy += 1

    #print "sx, sy =", sx, ", ", sy
    #print "start:", p_start
    #print "end:", p_end
    
    cdef int coord_s = p_start[0] + p_start[1]*sx
    cdef int coord_e = p_end[0] + p_end[1]*sx
    
    # At most this is 256x256, resulting in 64K char arrays
    # and 256K int arrays (assuming a char = 1 byte and int = 4 bytes)
    cdef char *s_valid   = <char*>malloc(sx*sy*sizeof(char))
    cdef char *s_cheap   = <char*>malloc(sx*sy*sizeof(char))
    cdef char *s_visited = <char*>malloc(sx*sy*sizeof(char))
    cdef int  *d_cost    = <int*>malloc(sx*sy*sizeof(int))
    
    cdef SortedList * to_check = new_slist()
    slist_add(to_check, coord_s, 0)
    
    for y in xrange(sy):
        for x in xrange(sx):
            s_valid[x + sx*y] = 1 if (x, y) in set_valid else 0
            s_cheap[x + sx*y] = 1 if (x, y) in set_cheap else 0
            s_visited[x + sx*y] = 0
            d_cost[x + sx*y] = 0
    
    cdef int nb[4]
    cdef int n, p, cost_p, p_x, p_y, cc
    cdef int min_c, min_p    

    #print "Begin..."
    
    while slist_is_empty(to_check) == 0:
        p      = slist_head_coord(to_check)
        cost_p = slist_head_cost(to_check)
        d_cost[p] = cost_p
        p_x = p % sx
        p_y = p // sx
        slist_pop(to_check)

        #print "Checking", p_x, ",", p_y, "(has cost: %i)" % cost_p
        
        s_visited[p] = 1

        if p == coord_e:
            #print "Found!"
            break

        nb[0] = -1 if p_x == 0 else p-1
        nb[1] = -1 if p_x >= sx-1 else p+1
        nb[2] = -1 if p_y == 0 else p-sx
        nb[3] = -1 if p_y >= sy-1 else p+sx
        
        for i in xrange(4):
            n = nb[i]
            if n == -1:
                continue
            if s_visited[n] == 1:
                continue
            if s_valid[n] == 0:
                continue

            if s_cheap[n] == 1:
                cc = cost_p + 1
            elif rand() % 100 <= 65:
                cc = cost_p + 4 + (rand() % 12)
            else:
                cc = cost_p + 4

            slist_add(to_check, n, cc)

    #print "done."

    # No route after all
    if s_visited[coord_e] == 0:
        rv = None

    # Backtrack otherwise
    else:
        # Display cost matrix
        #tmp = "Column--"
        #for x in xrange(sx):
        #    tmp += "%03u-" % x
        #print tmp
        #for y in xrange(sy):
        #    tmp = "Row%03u: " % y
        #    for x in xrange(sx):
        #        if (x, y) == p_start:
        #            tmp += "*S*"
        #        elif (x, y) == p_end:
        #            tmp += "*E*"
        #        else:
        #            tmp += "%03u" % d_cost[x + sx*y]
        #        tmp += " "
        #    print tmp

        tmp = [coord_e]
        while tmp[-1] != coord_s and len(tmp) <= sx*sy:
            p = tmp[-1]
            p_x = p % sx
            p_y = p // sx

            nb[0] = -1 if p_x == 0 else p-1
            nb[1] = -1 if p_x >= sx-1 else p+1
            nb[2] = -1 if p_y == 0 else p-sx
            nb[3] = -1 if p_y >= sy-1 else p+sx

            min_c = -1
            min_p = -1
            for i in xrange(4):
                n = nb[i]
                if n < 0:
                    continue
                if s_visited[n] == 0:
                    continue
                if min_c < 0 or min_c > d_cost[n]:
                    min_c = d_cost[n]
                    min_p = n
            #print p_x, p_y, "->", min_p % sx, min_p // sx
            assert min_c >= 0
            tmp.append(min_p)
    
            rv = []
            for v in tmp:
                rv.append((v % sx, v // sx))

        if tmp[-1] != coord_s:
            raise ValueError, "Could not backtrack found route!"

    free(<void*>s_valid)
    free(<void*>s_cheap)
    free(<void*>s_visited)
    free(<void*>d_cost)
    delete_slist(to_check)

    return rv
