#!/usr/bin/env python
#------------------------------------------------------------------------------
#    Copyright (C) 2009 Florian Schanda <florian@schanda.org.uk>
#
#    This file is part of the Warzone Map Tools.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program. If not, see <http://www.gnu.org/licenses/>.
#------------------------------------------------------------------------------

import random

class TileSet:
    def __init__(self):
        self.terrains = []
        self.terraintiles = []
        self.transtiles = []
        self.default = None
        self.cliff_x = None

    def get_alltrans(self, a):
        rv = []
        for t in self.transtiles.keys():
            if t[0] == a:
                tmp = t[1]
            elif t[1] == a:
                tmp = t[0]
            else:
                continue
            if tmp in self.terrains:
                rv.append(tmp)
        if len(rv) == 0:
            print "Could not find transition for '%s'." % a
        return rv
        
        
    def get_possible_transition(self, a):
        """ Given a base terrain a, returns a random terrain to transition to.
        """
        alltrans = self.get_alltrans(a)
        if not alltrans:
            return None
        if len(alltrans) == 1:
            return alltrans[0]

        # This makes it more likely to use all possible tiles
        tmp = random.choice(alltrans)
        if tmp == self.default:
            tmp = random.choice(alltrans)
        return tmp
            

    def getname(self, t):
        """ Given a random flat tile t, return the name. """
        for k in self.terraintiles:
            if t in (x[0] for x in self.terraintiles[k]):
                return k
        return None

    def getrandom(self, t):
        """ Given a terrain type t, returns a random tile of that type.
        """
        rv = {}
        tt = self.terraintiles[t]
        if len(tt) > 1:
            tmp = sum(x[1] for x in tt)
            n = random.randint(1, tmp)
            i = 0
            for t in tt:
                i += t[1]
                if i >= n:
                    rv["tex"] = t[0]
                    break
            assert(rv.get("tex", None) != None)
        else:
            rv["tex"] = tt[0][0]
        rv["rot"] = random.randint(0, 3)
        rv["flipx"] = bool(random.randint(0, 1))
        rv["flipy"] = bool(random.randint(0, 1))
        return rv

    def trunk_get_tile(self, t):
        """ TRUNK ONLY: Gets the most common transition.
        """
        tt = sorted(self.terraintiles[t], key=lambda x: tuple(reversed(x)))
        rv = {"tex": tt[-1][0], "rot": 0, "flipx": False, "flipy": False }
        return rv

    def rot(self, l):
        """ Given a transition layout l, returns it rotated by 90 deg to
            the right.
        """
        return (l[2], l[0],
                l[3], l[1])

    def rot_road(self, l):
        return (      l[1],
                l[3],       l[0],
                      l[2])

    def gettrans(self, a, b, layout):
        """ Given two terrain types a, b and a transition layout l,
            returns suitable transition tile.
        """
        if type(layout) is not tuple:
            layout = tuple(layout)

        # Deal with roads
        if "Road" in (a, b) or "Track" in (a, b):
            rf = self.rot_road
        else:
            rf = self.rot

        # print a, "->", b, layout
        
        # Deal with reverse transitions
        if (a, b) in self.transtiles:
            trans = self.transtiles[(a, b)]
        elif (b, a) in self.transtiles:
            trans = self.transtiles[(b, a)]
            layout = tuple(map(lambda x: int(not x), layout))

        # Find transition
        tmp = filter(lambda x: sum(x) == sum(layout), trans)
        if len(tmp) == 0:
            if layout == (1, 1, 1, 1):
                return self.getrandom(b)
            elif layout == (0, 0, 0, 0):
                return self.getrandom(a)
            
            raise ValueError, "Could not find suitable transition from '%s' " \
                "to '%s' for layout %s." % (a, b, layout)

        # Find rotation
        for t in tmp:
            rv = {"tex": trans[t]}
            if type(rv["tex"]) == tuple:
                rv = {"tex": random.choice(rv["tex"]) }

            for n in xrange(4):
                if layout == t:
                    return rv
                rv["rot"] = rv.get("rot", 0) + 1
                t = rf(t)
        
        print "Warning:", a, "->", b, "(layout:", layout, "): No such tile."
        return None

    def simple_texpath(self, a, b):
        if a in self.get_alltrans(b):
            return None
        return (set(self.get_alltrans(a)) &
                set(self.get_alltrans(b)))

    def get_default(self):
        if not self.default:
            possibles = set(self.get_alltrans("Cliff"))
            for p in set(possibles):
                for n in ("Road", "Track"):
                    if len(self.get_alltrans(n)):
                        tp = self.simple_texpath(p, n)
                        if tp is None or len(tp):
                            continue
                        possibles.remove(p)
            
            self.default = random.choice(list(possibles))
            print "Default set to %s." % self.default
        return self.default

    def get_set(self, name):
        """ This returns (as a frozenset) all tiles which are considered
            to be part of the terrain type 'name'.
        """
        rv = set(x[0] for x in self.terraintiles[name])
        for x in (list(v.itervalues())
                  for k, v in self.transtiles.iteritems()
                  if name in k):
            for t in x:
                if type(t) == tuple:
                    rv |= set(t)
                else:
                    rv.add(t)
        return frozenset(rv)

def get_random_tileset():
    n = random.randint(0, 2)
    if n == 0:
        import arizona
        return arizona.Arizona()
    elif n == 1:
        import urban
        return urban.Urban()
    elif n == 2:
        import rockies
        return rockies.Rockies()

def get_tileset(name):
    if name == "Arizona":
        import arizona
        return arizona.Arizona()
    elif name == "Urban":
        import urban
        return urban.Urban()
    elif name == "Rockies":
        import rockies
        return rockies.Rockies()
    else:
        return get_random_tileset()
